from functools import lru_cache
import logging
import numpy as np
from datetime import datetime
from dataclasses import dataclass
from typing import Optional, List, Any, cast
from numpy.typing import NDArray

from opentrons import config

from opentrons.config.robot_configs import (
    get_legacy_gantry_calibration,
    default_ot2_deck_calibration,
)
from opentrons.calibration_storage import (
    types,
    save_robot_deck_attitude,
    get_robot_deck_attitude,
)
from opentrons.util import linal

from .util import DeckTransformState

log = logging.getLogger(__name__)


@dataclass
class DeckCalibration:
    attitude: types.AttitudeMatrix
    source: types.SourceType
    status: types.CalibrationStatus
    belt_attitude: Optional[types.AttitudeMatrix] = None
    last_modified: Optional[datetime] = None
    pipette_calibrated_with: Optional[str] = None
    tiprack: Optional[str] = None


@dataclass
class RobotCalibration:
    deck_calibration: DeckCalibration


def validate_attitude_deck_calibration(
    deck_cal: DeckCalibration,
) -> DeckTransformState:
    """
    This function determines whether the deck calibration is valid
    or not based on the following use-cases:

    TODO(lc, 8/10/2020): As with the OT2, expand on this method, or create
    another method to diagnose bad instrument offset data
    """
    curr_cal: linal.DoubleMatrix = np.array(deck_cal.attitude)
    row, _ = curr_cal.shape
    rank: int = np.linalg.matrix_rank(curr_cal)
    if row != rank:
        # Check that the matrix is non-singular
        return DeckTransformState.SINGULARITY
    elif not deck_cal.last_modified:
        # Check that the matrix is not an identity
        return DeckTransformState.IDENTITY
    else:
        # Transform as it stands is sufficient.
        return DeckTransformState.OK


def validate_gantry_calibration(gantry_cal: List[List[float]]) -> DeckTransformState:
    """
    This function determines whether the gantry calibration is valid
    or not based on the following use-cases:
    """
    curr_cal: linal.DoubleMatrix = np.array(gantry_cal)
    row, _ = curr_cal.shape

    rank: int = np.linalg.matrix_rank(curr_cal)

    id_matrix = linal.identity_deck_transform()

    z = abs(curr_cal[2][-1])

    outofrange = z < 16 or z > 34
    if row != rank:
        # Check that the matrix is non-singular
        return DeckTransformState.SINGULARITY
    elif np.array_equal(curr_cal, id_matrix):
        # Check that the matrix is not an identity
        return DeckTransformState.IDENTITY
    elif outofrange:
        # Check that the matrix is not out of range.
        return DeckTransformState.BAD_CALIBRATION
    else:
        # Transform as it stands is sufficient.
        return DeckTransformState.OK


def migrate_affine_xy_to_attitude(
    gantry_cal: List[List[float]],
) -> types.AttitudeMatrix:
    masked_transform: NDArray[np.bool_] = np.array(
        [
            [True, True, True, False],
            [True, True, True, False],
            [False, False, False, False],
            [False, False, False, False],
        ]
    )
    masked_array: np.ma.MaskedArray[
        Any, np.dtype[np.float64]
    ] = np.ma.masked_array(  # type: ignore
        gantry_cal, ~masked_transform
    )
    attitude_array: linal.DoubleMatrix = np.zeros((3, 3))
    np.put(attitude_array, [0, 1, 2], masked_array[0].compressed())
    np.put(attitude_array, [3, 4, 5], masked_array[1].compressed())
    np.put(attitude_array, 8, 1)
    return cast(List[List[float]], attitude_array.tolist())


def save_attitude_matrix(
    expected: linal.SolvePoints,
    actual: linal.SolvePoints,
    pipette_id: str,
    tiprack_hash: str,
) -> None:
    attitude = linal.solve_attitude(expected, actual)
    save_robot_deck_attitude(
        attitude,
        pipette_id,
        tiprack_hash,
    )


def load_attitude_matrix() -> DeckCalibration:
    calibration_data = get_robot_deck_attitude()
    gantry_cal = get_legacy_gantry_calibration()
    if not calibration_data and gantry_cal:
        if validate_gantry_calibration(gantry_cal) == DeckTransformState.OK:
            log.debug(
                "Attitude deck calibration matrix not found. Migrating "
                "existing affine deck calibration matrix to {}".format(
                    config.get_opentrons_path("robot_calibration_dir")
                )
            )
            attitude = migrate_affine_xy_to_attitude(gantry_cal)
            save_robot_deck_attitude(
                transform=attitude,
                pip_id=None,
                lw_hash=None,
                source=types.SourceType.legacy,
            )
            calibration_data = get_robot_deck_attitude()

    if calibration_data:
        return DeckCalibration(
            attitude=calibration_data.attitude,
            source=calibration_data.source,
            status=types.CalibrationStatus(**calibration_data.status.model_dump()),
            last_modified=calibration_data.last_modified,
            pipette_calibrated_with=calibration_data.pipette_calibrated_with,
            tiprack=calibration_data.tiprack,
        )
    else:
        # load default if deck calibration data do not exist
        return DeckCalibration(
            attitude=default_ot2_deck_calibration(),
            source=types.SourceType.default,
            status=types.CalibrationStatus(),
        )


def load() -> RobotCalibration:
    return RobotCalibration(deck_calibration=load_attitude_matrix())


class RobotCalibrationProvider:
    def __init__(self) -> None:
        self._robot_calibration = load()

    @lru_cache(1)
    def _validate(self) -> DeckTransformState:
        return validate_attitude_deck_calibration(
            self._robot_calibration.deck_calibration
        )

    @property
    def robot_calibration(self) -> RobotCalibration:
        return self._robot_calibration

    def reset_robot_calibration(self) -> None:
        self._validate.cache_clear()
        self._robot_calibration = load()

    def reset_deck_calibration(self) -> None:
        self._robot_calibration = load()

    def load_deck_calibration(self) -> None:
        self._robot_calibration = load()

    def set_robot_calibration(self, robot_calibration: RobotCalibration) -> None:
        self._validate.cache_clear()
        self._robot_calibration = robot_calibration

    def validate_calibration(self) -> DeckTransformState:
        """
        The lru cache decorator is currently not supported by the
        ThreadManager. To work around this, we need to wrap the
        actual function around a dummy outer function.

        Once decorators are more fully supported, we can remove this.
        """
        return self._validate()

    def build_temporary_identity_calibration(self) -> RobotCalibration:
        """
        Get a temporary identity deck cal suitable for use during
        calibration
        """
        return RobotCalibration(
            deck_calibration=DeckCalibration(
                attitude=default_ot2_deck_calibration(),
                source=types.SourceType.default,
                status=types.CalibrationStatus(),
            )
        )



"""Basic labware data state and store."""

from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    NamedTuple,
    Union,
    overload,
)
from typing_extensions import assert_never

from opentrons.protocol_engine.state import update_types
from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.gripper.constants import LABWARE_GRIP_FORCE
from opentrons_shared_data.labware.labware_definition import (
    InnerWellGeometry,
    LabwareDefinition,
    LabwareRole,
    WellDefinition2,
    WellDefinition3,
)
from opentrons_shared_data.pipette.types import LabwareUri

from opentrons.types import DeckSlotName, StagingSlotName, MountType
from opentrons.protocols.api_support.constants import OPENTRONS_NAMESPACE
from opentrons.calibration_storage.helpers import uri_from_details

from .. import errors
from ..resources import DeckFixedLabware, labware_validation, fixture_validation
from ..types import (
    DeckSlotLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    NonStackedLocation,
    Dimensions,
    LabwareOffset,
    LabwareOffsetVector,
    LabwareOffsetLocationSequence,
    LegacyLabwareOffsetLocation,
    InStackerHopperLocation,
    LabwareLocation,
    LoadedLabware,
    ModuleLocation,
    ModuleModel,
    OverlapOffset,
    LabwareMovementOffsetData,
    OnDeckLabwareLocation,
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
)
from ..actions import (
    Action,
    AddLabwareOffsetAction,
    AddLabwareDefinitionAction,
    get_state_updates,
)
from ._abstract_store import HasState, HandlesActions
from ._move_types import EdgePathType


# URIs of labware whose definitions accidentally specify an engage height
# in units of half-millimeters instead of millimeters.
_MAGDECK_HALF_MM_LABWARE = {
    "opentrons/biorad_96_wellplate_200ul_pcr/1",
    "opentrons/nest_96_wellplate_100ul_pcr_full_skirt/1",
    "opentrons/usascientific_96_wellplate_2.4ml_deep/1",
}

_RIGHT_SIDE_SLOTS = {
    # OT-2:
    DeckSlotName.FIXED_TRASH,
    DeckSlotName.SLOT_9,
    DeckSlotName.SLOT_6,
    DeckSlotName.SLOT_3,
    # OT-3:
    DeckSlotName.SLOT_A3,
    DeckSlotName.SLOT_B3,
    DeckSlotName.SLOT_C3,
    DeckSlotName.SLOT_D3,
}


# The max height of the labware that can fit in a plate reader
_PLATE_READER_MAX_LABWARE_Z_MM = 16


_WellDefinition = WellDefinition2 | WellDefinition3


class LabwareLoadParams(NamedTuple):
    """Parameters required to load a labware in Protocol Engine."""

    load_name: str
    namespace: str
    version: int


@dataclass
class LabwareState:
    """State of all loaded labware resources."""

    # Indexed by LoadedLabware.id.
    # If a LoadedLabware here has a non-None offsetId,
    # it must point to an existing element of labware_offsets_by_id.
    labware_by_id: Dict[str, LoadedLabware]

    # Indexed by LabwareOffset.id.
    # We rely on Python 3.7+ preservation of dict insertion order.
    labware_offsets_by_id: Dict[str, LabwareOffset]

    definitions_by_uri: Dict[str, LabwareDefinition]
    deck_definition: DeckDefinitionV5


class LabwareStore(HasState[LabwareState], HandlesActions):
    """Labware state container."""

    _state: LabwareState

    def __init__(
        self,
        deck_definition: DeckDefinitionV5,
        deck_fixed_labware: Sequence[DeckFixedLabware],
    ) -> None:
        """Initialize a labware store and its state."""
        definitions_by_uri: Dict[str, LabwareDefinition] = {
            uri_from_details(
                load_name=fixed_labware.definition.parameters.loadName,
                namespace=fixed_labware.definition.namespace,
                version=fixed_labware.definition.version,
            ): fixed_labware.definition
            for fixed_labware in deck_fixed_labware
        }
        labware_by_id = {
            fixed_labware.labware_id: LoadedLabware.model_construct(
                id=fixed_labware.labware_id,
                location=fixed_labware.location,
                loadName=fixed_labware.definition.parameters.loadName,
                definitionUri=uri_from_details(
                    load_name=fixed_labware.definition.parameters.loadName,
                    namespace=fixed_labware.definition.namespace,
                    version=fixed_labware.definition.version,
                ),
                offsetId=None,
            )
            for fixed_labware in deck_fixed_labware
        }

        self._state = LabwareState(
            definitions_by_uri=definitions_by_uri,
            labware_offsets_by_id={},
            labware_by_id=labware_by_id,
            deck_definition=deck_definition,
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._add_loaded_labware(state_update)
            self._add_batch_loaded_labwares(state_update)
            self._add_loaded_lid_stack(state_update)
            self._set_labware_location(state_update)
            self._set_batch_labware_location(state_update)
            self._set_labware_lid(state_update)

        if isinstance(action, AddLabwareOffsetAction):
            labware_offset = LabwareOffset.model_construct(
                id=action.labware_offset_id,
                createdAt=action.created_at,
                definitionUri=action.request.definitionUri,
                location=action.request.legacyLocation,
                locationSequence=action.request.locationSequence,
                vector=action.request.vector,
            )
            self._add_labware_offset(labware_offset)

        elif isinstance(action, AddLabwareDefinitionAction):
            uri = uri_from_details(
                namespace=action.definition.namespace,
                load_name=action.definition.parameters.loadName,
                version=action.definition.version,
            )
            self._state.definitions_by_uri[uri] = action.definition

    def _add_labware_offset(self, labware_offset: LabwareOffset) -> None:
        """Add a new labware offset to state.

        `labware_offset.id` must not match any existing labware offset ID.
        `LoadLabwareCommand`s retain references to their corresponding labware offsets
        and expect them to be immutable.
        """
        assert labware_offset.id not in self._state.labware_offsets_by_id

        self._state.labware_offsets_by_id[labware_offset.id] = labware_offset

    def _add_loaded_labware(self, state_update: update_types.StateUpdate) -> None:
        loaded_labware_update = state_update.loaded_labware
        if loaded_labware_update != update_types.NO_CHANGE:
            # If the labware load refers to an offset, that offset must actually exist.
            if loaded_labware_update.offset_id is not None:
                assert (
                    loaded_labware_update.offset_id in self._state.labware_offsets_by_id
                )

            definition_uri = uri_from_details(
                namespace=loaded_labware_update.definition.namespace,
                load_name=loaded_labware_update.definition.parameters.loadName,
                version=loaded_labware_update.definition.version,
            )

            self._state.definitions_by_uri[
                definition_uri
            ] = loaded_labware_update.definition

            location = loaded_labware_update.new_location

            display_name = loaded_labware_update.display_name

            self._state.labware_by_id[
                loaded_labware_update.labware_id
            ] = LoadedLabware.model_construct(
                id=loaded_labware_update.labware_id,
                location=location,
                loadName=loaded_labware_update.definition.parameters.loadName,
                definitionUri=definition_uri,
                offsetId=loaded_labware_update.offset_id,
                displayName=display_name,
            )

    def _add_batch_loaded_labwares(
        self, state_update: update_types.StateUpdate
    ) -> None:
        batch_loaded_labware_update = state_update.batch_loaded_labware
        if batch_loaded_labware_update == update_types.NO_CHANGE:
            return
        # If the labware load refers to an offset, that offset must actually exist.
        for labware_id in batch_loaded_labware_update.new_locations_by_id:
            if batch_loaded_labware_update.offset_ids_by_id[labware_id] is not None:
                assert (
                    batch_loaded_labware_update.offset_ids_by_id[labware_id]
                    in self._state.labware_offsets_by_id
                )

            definition_uri = uri_from_details(
                namespace=batch_loaded_labware_update.definitions_by_id[
                    labware_id
                ].namespace,
                load_name=batch_loaded_labware_update.definitions_by_id[
                    labware_id
                ].parameters.loadName,
                version=batch_loaded_labware_update.definitions_by_id[
                    labware_id
                ].version,
            )

            self._state.definitions_by_uri[
                definition_uri
            ] = batch_loaded_labware_update.definitions_by_id[labware_id]

            location = batch_loaded_labware_update.new_locations_by_id[labware_id]

            self._state.labware_by_id[labware_id] = LoadedLabware.model_construct(
                id=labware_id,
                location=location,
                loadName=batch_loaded_labware_update.definitions_by_id[
                    labware_id
                ].parameters.loadName,
                definitionUri=definition_uri,
                offsetId=batch_loaded_labware_update.offset_ids_by_id[labware_id],
                displayName=batch_loaded_labware_update.display_names_by_id[labware_id],
            )

    def _add_loaded_lid_stack(self, state_update: update_types.StateUpdate) -> None:
        loaded_lid_stack_update = state_update.loaded_lid_stack
        if loaded_lid_stack_update != update_types.NO_CHANGE:
            # Add the stack object
            stack_definition_uri = uri_from_details(
                namespace=loaded_lid_stack_update.stack_object_definition.namespace,
                load_name=loaded_lid_stack_update.stack_object_definition.parameters.loadName,
                version=loaded_lid_stack_update.stack_object_definition.version,
            )
            self.state.definitions_by_uri[
                stack_definition_uri
            ] = loaded_lid_stack_update.stack_object_definition
            self._state.labware_by_id[
                loaded_lid_stack_update.stack_id
            ] = LoadedLabware.construct(
                id=loaded_lid_stack_update.stack_id,
                location=loaded_lid_stack_update.stack_location,
                loadName=loaded_lid_stack_update.stack_object_definition.parameters.loadName,
                definitionUri=stack_definition_uri,
                offsetId=None,
                displayName=None,
            )

            # Add the Lids on top of the stack object
            for labware_id in loaded_lid_stack_update.new_locations_by_id:
                if loaded_lid_stack_update.definition is None:
                    raise ValueError(
                        "Lid Stack Labware Definition cannot be None when multiple lids are loaded."
                    )
                definition_uri = uri_from_details(
                    namespace=loaded_lid_stack_update.definition.namespace,
                    load_name=loaded_lid_stack_update.definition.parameters.loadName,
                    version=loaded_lid_stack_update.definition.version,
                )

                self._state.definitions_by_uri[
                    definition_uri
                ] = loaded_lid_stack_update.definition

                location = loaded_lid_stack_update.new_locations_by_id[labware_id]

                self._state.labware_by_id[labware_id] = LoadedLabware.construct(
                    id=labware_id,
                    location=location,
                    loadName=loaded_lid_stack_update.definition.parameters.loadName,
                    definitionUri=definition_uri,
                    offsetId=None,
                    displayName=None,
                )

    def _set_labware_lid(self, state_update: update_types.StateUpdate) -> None:
        labware_lid_update = state_update.labware_lid
        if labware_lid_update != update_types.NO_CHANGE:
            parent_labware_ids = labware_lid_update.parent_labware_ids
            for i in range(len(parent_labware_ids)):
                lid_id = labware_lid_update.lid_ids[i]
                self._state.labware_by_id[parent_labware_ids[i]].lid_id = lid_id

    def _do_update_labware_location(
        self, labware_id: str, new_location: LabwareLocation, new_offset_id: str | None
    ) -> None:
        self._state.labware_by_id[labware_id].offsetId = new_offset_id

        if isinstance(new_location, AddressableAreaLocation) and (
            fixture_validation.is_gripper_waste_chute(new_location.addressableAreaName)
            or fixture_validation.is_trash(new_location.addressableAreaName)
        ):
            # If a labware has been moved into a waste chute it's been chuted away and is now technically off deck
            new_location = OFF_DECK_LOCATION

        self._state.labware_by_id[labware_id].location = new_location

    def _set_labware_location(self, state_update: update_types.StateUpdate) -> None:
        labware_location_update = state_update.labware_location
        if labware_location_update == update_types.NO_CHANGE:
            return

        self._do_update_labware_location(
            labware_location_update.labware_id,
            labware_location_update.new_location,
            labware_location_update.offset_id,
        )

    def _set_batch_labware_location(
        self, state_update: update_types.StateUpdate
    ) -> None:
        batch_location_update = state_update.batch_labware_location
        if batch_location_update == update_types.NO_CHANGE:
            return
        for (
            labware_id,
            new_location,
        ) in batch_location_update.new_locations_by_id.items():
            self._do_update_labware_location(
                labware_id,
                new_location,
                batch_location_update.new_offset_ids_by_id.get(labware_id, None),
            )


class LabwareView:
    """Read-only labware state view."""

    _state: LabwareState

    def __init__(self, state: LabwareState) -> None:
        """Initialize the computed view of labware state.

        Arguments:
            state: Labware state dataclass used for all calculations.
        """
        self._state = state

    def get(self, labware_id: str) -> LoadedLabware:
        """Get labware data by the labware's unique identifier."""
        try:
            return self._state.labware_by_id[labware_id]
        except KeyError as e:
            raise errors.LabwareNotLoadedError(
                f"Labware {labware_id} not found."
            ) from e

    def get_id_by_module(self, module_id: str) -> str:
        """Return the ID of the labware loaded on the given module."""
        for labware_id, labware in self._state.labware_by_id.items():
            if (
                isinstance(labware.location, ModuleLocation)
                and labware.location.moduleId == module_id
            ):
                return labware_id

        raise errors.exceptions.LabwareNotLoadedOnModuleError(
            "There is no labware loaded on this Module"
        )

    def get_id_by_labware(self, labware_id: str) -> str:
        """Return the ID of the labware loaded on the given labware."""
        for labware in self._state.labware_by_id.values():
            if (
                isinstance(labware.location, OnLabwareLocation)
                and labware.location.labwareId == labware_id
            ):
                return labware.id
        raise errors.exceptions.LabwareNotLoadedOnLabwareError(
            f"There is not labware loaded onto labware {labware_id}"
        )

    def raise_if_labware_has_non_lid_labware_on_top(self, labware_id: str) -> None:
        """Raise if labware has another labware that is not its lid on top."""
        lid_id = self.get_lid_id_by_labware_id(labware_id)
        for candidate_id, candidate_labware in self._state.labware_by_id.items():
            if (
                isinstance(candidate_labware.location, OnLabwareLocation)
                and candidate_labware.location.labwareId == labware_id
                and candidate_id != lid_id
            ):
                raise errors.LabwareIsInStackError(
                    f"Cannot access labware {labware_id} because it has a non-lid labware stacked on top."
                )

    def raise_if_labware_has_labware_on_top(self, labware_id: str) -> None:
        """Raise if labware has another labware on top."""
        for labware in self._state.labware_by_id.values():
            if (
                isinstance(labware.location, OnLabwareLocation)
                and labware.location.labwareId == labware_id
            ):
                raise errors.LabwareIsInStackError(
                    f"Cannot access labware {labware_id} because it has another labware stacked on top."
                )

    def get_by_slot(
        self,
        slot_name: Union[DeckSlotName, StagingSlotName],
    ) -> Optional[LoadedLabware]:
        """Get the labware located in a given slot, if any."""
        loaded_labware = list(self._state.labware_by_id.values())

        for labware in loaded_labware:
            if (
                isinstance(labware.location, DeckSlotLocation)
                and labware.location.slotName.id == slot_name.id
            ) or (
                isinstance(labware.location, AddressableAreaLocation)
                and labware.location.addressableAreaName == slot_name.id
            ):
                return labware

        return None

    def get_by_addressable_area(
        self,
        addressable_area: str,
    ) -> Optional[LoadedLabware]:
        """Get the labware located in a given addressable area, if any."""
        loaded_labware = list(self._state.labware_by_id.values())

        for labware in loaded_labware:
            if (
                isinstance(labware.location, AddressableAreaLocation)
                and labware.location.addressableAreaName == addressable_area
            ):
                return labware

        return None

    def get_definition(self, labware_id: str) -> LabwareDefinition:
        """Get labware definition by the labware's unique identifier."""
        return self.get_definition_by_uri(
            LabwareUri(self.get(labware_id).definitionUri)
        )

    def get_user_specified_display_name(self, labware_id: str) -> Optional[str]:
        """Get the labware's user-specified display name, if set."""
        return self.get(labware_id).displayName

    def get_display_name(self, labware_id: str) -> str:
        """Get the labware's display name.

        If a user-specified display name exists, will return that, else will return
        display name from the definition.
        """
        return (
            self.get_user_specified_display_name(labware_id)
            or self.get_definition(labware_id).metadata.displayName
        )

    def get_deck_definition(self) -> DeckDefinitionV5:
        """Get the current deck definition."""
        return self._state.deck_definition

    def get_definition_by_uri(self, uri: LabwareUri) -> LabwareDefinition:
        """Get the labware definition matching loadName namespace and version."""
        try:
            return self._state.definitions_by_uri[uri]
        except KeyError as e:
            raise errors.LabwareDefinitionDoesNotExistError(
                f"Labware definition for matching {uri} not found."
            ) from e

    def get_loaded_labware_definitions(self) -> List[LabwareDefinition]:
        """Get all loaded labware definitions."""
        loaded_labware = self._state.labware_by_id.values()
        return [
            self.get_definition_by_uri(LabwareUri(labware.definitionUri))
            for labware in loaded_labware
        ]

    def find_custom_labware_load_params(self) -> List[LabwareLoadParams]:
        """Find all load labware parameters for custom labware definitions in state."""
        return [
            LabwareLoadParams(
                load_name=definition.parameters.loadName,
                namespace=definition.namespace,
                version=definition.version,
            )
            for definition in self._state.definitions_by_uri.values()
            if definition.namespace != OPENTRONS_NAMESPACE
        ]

    def get_location(self, labware_id: str) -> LabwareLocation:
        """Get labware location by the labware's unique identifier."""
        return self.get(labware_id).location

    def get_parent_location(self, labware_id: str) -> NonStackedLocation:
        """Get labware's non-labware parent location."""
        parent = self.get_location(labware_id)
        if isinstance(parent, OnLabwareLocation):
            return self.get_parent_location(parent.labwareId)
        elif isinstance(parent, InStackerHopperLocation):
            # TODO: This function really wants to return something like an "EventuallyOnDeckLocation"
            # and either raise or return None for labware that isn't traceable to a place on the robot
            # deck (i.e. not in a stacker hopper, not off-deck, not in system). We don't really have
            # that concept yet but should add it soon. In the meantime, other checks should prevent
            # this being called in those cases.
            return ModuleLocation(moduleId=parent.moduleId)
        return parent

    def get_highest_child_labware(self, labware_id: str) -> str:
        """Get labware's highest child labware returning the labware ID."""
        if (child_id := self.get_next_child_labware(labware_id)) is not None:
            return self.get_highest_child_labware(labware_id=child_id)
        return labware_id

    def get_next_child_labware(self, labware_id: str) -> str | None:
        """Get the labware that is on this labware, if any.

        This includes lids.
        """
        for labware in self._state.labware_by_id.values():
            if (
                isinstance(labware.location, OnLabwareLocation)
                and labware.location.labwareId == labware_id
            ):
                return labware.id
        return None

    def get_labware_stack_from_parent(self, labware_id: str) -> list[str]:
        """Get the stack of labware starting from the specified labware ID and moving up."""
        labware_ids = [labware_id]
        while (next_id := self.get_next_child_labware(labware_id)) is not None:
            labware_ids.append(next_id)
            labware_id = next_id
        return labware_ids

    def get_labware_stack(
        self, labware_stack: List[LoadedLabware]
    ) -> List[LoadedLabware]:
        """Get the a stack of labware starting from a given labware or existing stack."""
        parent = self.get_location(labware_stack[-1].id)
        if isinstance(parent, OnLabwareLocation):
            labware_stack.append(self.get(parent.labwareId))
            return self.get_labware_stack(labware_stack)
        return labware_stack

    def get_lid_id_by_labware_id(self, labware_id: str) -> str | None:
        """Get the ID of a lid labware on top of a given labware, if any."""
        return self._state.labware_by_id[labware_id].lid_id

    def get_lid_by_labware_id(self, labware_id: str) -> LoadedLabware | None:
        """Get the Lid Labware that is currently on top of a given labware, if there is one."""
        lid_id = self.get_lid_id_by_labware_id(labware_id)
        if lid_id:
            return self._state.labware_by_id[lid_id]
        else:
            return None

    def get_labware_by_lid_id(self, lid_id: str) -> LoadedLabware | None:
        """Get the labware that is currently covered by a given lid, if there is one."""
        loaded_labware = list(self._state.labware_by_id.values())
        for labware in loaded_labware:
            if labware.lid_id == lid_id:
                return labware
        return None

    def get_all(self) -> List[LoadedLabware]:
        """Get a list of all labware entries in state."""
        return list(self._state.labware_by_id.values())

    def get_has_quirk(self, labware_id: str, quirk: str) -> bool:
        """Get if a labware has a certain quirk."""
        return quirk in self.get_quirks(labware_id=labware_id)

    def get_quirks(self, labware_id: str) -> List[str]:
        """Get a labware's quirks."""
        definition = self.get_definition(labware_id)
        return definition.parameters.quirks or []

    def get_should_center_column_on_target_well(self, labware_id: str) -> bool:
        """True if a pipette moving to this labware should center its active column on the target.

        This is true for labware that have wells spanning entire columns.
        """
        has_quirk = self.get_has_quirk(labware_id, "centerMultichannelOnWells")
        return has_quirk and (
            len(self.get_definition(labware_id).wells) > 1
            and len(self.get_definition(labware_id).wells) < 96
        )

    def get_labware_stacking_maximum(self, labware: LabwareDefinition) -> int:
        """Returns the maximum number of labware allowed in a stack for a given labware definition.

        If not defined within a labware, defaults to one.
        """
        return labware.stackLimit if labware.stackLimit is not None else 1

    def get_should_center_pipette_on_target_well(self, labware_id: str) -> bool:
        """True if a pipette moving to a well of this labware should center its body on the target.

        This is true for 1-well reservoirs no matter the pipette, and for large plates.
        """
        has_quirk = self.get_has_quirk(labware_id, "centerMultichannelOnWells")
        return has_quirk and (
            len(self.get_definition(labware_id).wells) == 1
            or len(self.get_definition(labware_id).wells) >= 96
        )

    def get_well_definition(
        self,
        labware_id: str,
        well_name: Optional[str] = None,
    ) -> WellDefinition2 | WellDefinition3:
        """Get a well's definition by labware and well name.

        If `well_name` is omitted, the first well in the labware
        will be used.
        """
        definition = self.get_definition(labware_id)
        if well_name is None:
            well_name = definition.ordering[0][0]

        try:
            return definition.wells[well_name]
        except KeyError as e:
            raise errors.WellDoesNotExistError(
                f"{well_name} does not exist in {labware_id}."
            ) from e

    def get_well_geometry(
        self, labware_id: str, well_name: Optional[str] = None
    ) -> InnerWellGeometry:
        """Get a well's inner geometry by labware and well name."""
        labware_def = self.get_definition(labware_id)
        if labware_def.innerLabwareGeometry is None:
            raise errors.IncompleteLabwareDefinitionError(
                message=f"No innerLabwareGeometry found in labware definition for labware_id: {labware_id}."
            )
        well_def = self.get_well_definition(labware_id, well_name)
        geometry_id = well_def.geometryDefinitionId
        if geometry_id is None:
            raise errors.IncompleteWellDefinitionError(
                message=f"No geometryDefinitionId found in well definition for well: {well_name} in labware_id: {labware_id}"
            )
        else:
            well_geometry = labware_def.innerLabwareGeometry.get(geometry_id)
            if well_geometry is None:
                raise errors.IncompleteLabwareDefinitionError(
                    message=f"No innerLabwareGeometry found in labware definition for well_id: {geometry_id} in labware_id: {labware_id}"
                )
            return well_geometry

    def get_well_size(
        self, labware_id: str, well_name: str
    ) -> Tuple[float, float, float]:
        """Get a well's size in x, y, z dimensions based on its shape.

        Args:
            labware_id: Labware identifier.
            well_name: Name of well in labware.

        Returns:
            A tuple of dimensions in x, y, and z. If well is circular,
            the x and y dimensions will both be set to the diameter.
        """
        well_definition = self.get_well_definition(labware_id, well_name)

        if well_definition.shape == "circular":
            x_size = y_size = well_definition.diameter
        elif well_definition.shape == "rectangular":
            x_size = well_definition.xDimension
            y_size = well_definition.yDimension
        else:
            assert_never(well_definition.shape)

        return x_size, y_size, well_definition.depth

    def get_well_radial_offsets(
        self, labware_id: str, well_name: str, radius_percentage: float
    ) -> Tuple[float, float]:
        """Get x and y radius offsets modified by radius percentage."""
        x_size, y_size, z_size = self.get_well_size(labware_id, well_name)
        return (x_size / 2.0) * radius_percentage, (y_size / 2.0) * radius_percentage

    def get_edge_path_type(
        self,
        labware_id: str,
        well_name: str,
        mount: MountType,
        labware_slot: DeckSlotName,
        next_to_module: bool,
    ) -> EdgePathType:
        """Get the recommended edge path type based on well column, labware position and any neighboring modules."""
        labware_definition = self.get_definition(labware_id)
        left_column = labware_definition.ordering[0]
        right_column = labware_definition.ordering[-1]

        left_path_criteria = mount is MountType.RIGHT and well_name in left_column
        right_path_criteria = mount is MountType.LEFT and well_name in right_column
        labware_right_side = labware_slot in _RIGHT_SIDE_SLOTS

        if left_path_criteria and (next_to_module or labware_right_side):
            return EdgePathType.LEFT
        elif right_path_criteria and next_to_module:
            return EdgePathType.RIGHT
        else:
            return EdgePathType.DEFAULT

    def validate_liquid_allowed_in_labware(
        self, labware_id: str, wells: Mapping[str, Any]
    ) -> List[str]:
        """Check if wells associated to a labware_id has well by name and that labware is not tiprack."""
        labware_definition = self.get_definition(labware_id)
        labware_wells = labware_definition.wells
        contains_wells = all(well_name in labware_wells for well_name in iter(wells))
        if labware_definition.parameters.isTiprack:
            raise errors.LabwareIsTipRackError(
                f"Given labware: {labware_id} is a tiprack. Can not load liquid."
            )
        if LabwareRole.adapter in labware_definition.allowedRoles:
            raise errors.LabwareIsAdapterError(
                f"Given labware: {labware_id} is an adapter. Can not load liquid."
            )
        if not contains_wells:
            raise errors.WellDoesNotExistError(
                f"Some of the supplied wells do not match the labwareId: {labware_id}."
            )
        return list(wells)

    def get_tip_length(self, labware_id: str, overlap: float = 0) -> float:
        """Get the nominal tip length of a tip rack."""
        definition = self.get_definition(labware_id)
        if definition.parameters.tipLength is None:
            raise errors.LabwareIsNotTipRackError(
                f"Labware {labware_id} has no tip length defined."
            )

        return definition.parameters.tipLength - overlap

    def get_tip_drop_z_offset(
        self, labware_id: str, length_scale: float, additional_offset: float
    ) -> float:
        """Get the tip drop offset from the top of the well."""
        tip_length = self.get_tip_length(labware_id)
        return -tip_length * length_scale + additional_offset

    def get_definition_uri(self, labware_id: str) -> LabwareUri:
        """Get a labware's definition URI."""
        return LabwareUri(self.get(labware_id).definitionUri)

    def get_uri_from_definition(
        self,
        labware_definition: LabwareDefinition,
    ) -> LabwareUri:
        """Get a definition URI from a full labware definition."""
        return uri_from_details(
            load_name=labware_definition.parameters.loadName,
            namespace=labware_definition.namespace,
            version=labware_definition.version,
        )

    def is_tiprack(self, labware_id: str) -> bool:
        """Get whether labware is a tiprack."""
        definition = self.get_definition(labware_id)
        return definition.parameters.isTiprack

    def get_load_name(self, labware_id: str) -> str:
        """Get the labware's load name."""
        definition = self.get_definition(labware_id)
        return definition.parameters.loadName

    @overload
    def get_dimensions(self, *, labware_definition: LabwareDefinition) -> Dimensions:
        pass

    @overload
    def get_dimensions(self, *, labware_id: str) -> Dimensions:
        pass

    def get_dimensions(
        self,
        *,
        labware_definition: LabwareDefinition | None = None,
        labware_id: str | None = None,
    ) -> Dimensions:
        """Get the labware's dimensions."""
        if labware_definition is None:
            assert labware_id is not None  # From our @overloads.
            labware_definition = self.get_definition(labware_id)

        dims = labware_definition.dimensions

        return Dimensions(
            x=dims.xDimension,
            y=dims.yDimension,
            z=dims.zDimension,
        )

    def get_labware_overlap_offsets(
        self, definition: LabwareDefinition, below_labware_name: str
    ) -> OverlapOffset:
        """Get the labware's overlap with requested labware's load name."""
        if below_labware_name in definition.stackingOffsetWithLabware.keys():
            stacking_overlap = definition.stackingOffsetWithLabware.get(
                below_labware_name, OverlapOffset(x=0, y=0, z=0)
            )
        else:
            stacking_overlap = definition.stackingOffsetWithLabware.get(
                "default", OverlapOffset(x=0, y=0, z=0)
            )
        return OverlapOffset(
            x=stacking_overlap.x, y=stacking_overlap.y, z=stacking_overlap.z
        )

    def get_module_overlap_offsets(
        self, definition: LabwareDefinition, module_model: ModuleModel
    ) -> OverlapOffset:
        """Get the labware's overlap with requested module model."""
        stacking_overlap = definition.stackingOffsetWithModule.get(
            str(module_model.value)
        )
        if not stacking_overlap:
            if self._is_thermocycler_on_ot2(module_model):
                return OverlapOffset(x=0, y=0, z=10.7)
            else:
                return OverlapOffset(x=0, y=0, z=0)

        return OverlapOffset(
            x=stacking_overlap.x, y=stacking_overlap.y, z=stacking_overlap.z
        )

    def _is_thermocycler_on_ot2(self, module_model: ModuleModel) -> bool:
        """Whether the given module is a thermocycler with the current deck being an OT2 deck."""
        robot_model = self.get_deck_definition()["robot"]["model"]
        return (
            module_model
            in [ModuleModel.THERMOCYCLER_MODULE_V1, ModuleModel.THERMOCYCLER_MODULE_V2]
            and robot_model == "OT-2 Standard"
        )

    def get_default_magnet_height(self, module_id: str, offset: float) -> float:
        """Return a labware's default Magnetic Module engage height with added offset, if supplied.

        The returned value is measured in millimeters above the labware base plane.
        """
        labware_id = self.get_id_by_module(module_id)
        parameters = self.get_definition(labware_id).parameters
        default_engage_height = parameters.magneticModuleEngageHeight
        if (
            parameters.isMagneticModuleCompatible is False
            or default_engage_height is None
        ):
            raise errors.exceptions.NoMagnetEngageHeightError(
                "The labware loaded on this Magnetic Module"
                " does not have a default engage height."
            )

        if self._is_magnetic_module_uri_in_half_millimeter(labware_id):
            # TODO(mc, 2022-09-26): this value likely _also_ needs a few mm subtracted
            # https://opentrons.atlassian.net/browse/RSS-111
            calculated_height = default_engage_height / 2.0
        else:
            calculated_height = default_engage_height

        return calculated_height + offset

    def get_labware_offset_vector(self, labware_id: str) -> LabwareOffsetVector:
        """Get the labware's calibration offset."""
        offset_id = self.get(labware_id=labware_id).offsetId
        if offset_id is None:
            return LabwareOffsetVector(x=0, y=0, z=0)
        else:
            return self._state.labware_offsets_by_id[offset_id].vector

    def get_labware_offset(self, labware_offset_id: str) -> LabwareOffset:
        """Get a labware offset by the offset's unique ID.

        Raises:
            LabwareOffsetDoesNotExistError: If the given ID does not match any
                                            previously added offset.
        """
        try:
            return self._state.labware_offsets_by_id[labware_offset_id]
        except KeyError as e:
            raise errors.LabwareOffsetDoesNotExistError(
                f"Labware offset {labware_offset_id} not found."
            ) from e

    def get_labware_offsets(self) -> List[LabwareOffset]:
        """Get all labware offsets, in the order they were added."""
        return list(self._state.labware_offsets_by_id.values())

    def find_applicable_labware_offset(
        self, definition_uri: str, location: LabwareOffsetLocationSequence
    ) -> Optional[LabwareOffset]:
        """Find a labware offset that applies to the given definition and location sequence.

        Returns the *most recently* added matching offset, so later ones can override earlier ones.
        Returns ``None`` if no loaded offset matches the location.

        An offset matches a labware instance if the sequence of locations formed by following the
        .location elements of the labware instance until you reach an addressable area has the same
        definition URIs as the sequence of definition URIs stored by the offset.
        """
        for candidate in reversed(list(self._state.labware_offsets_by_id.values())):
            if (
                candidate.definitionUri == definition_uri
                and candidate.locationSequence == location
            ):
                return candidate
        return None

    def find_applicable_labware_offset_by_legacy_location(
        self,
        definition_uri: str,
        location: LegacyLabwareOffsetLocation,
    ) -> Optional[LabwareOffset]:
        """Find a labware offset that applies to the given definition and legacy location.

        Returns the *most recently* added matching offset,
        so later offsets can override earlier ones.
        Or, ``None`` if no offsets match at all.

        An offset "matches"
        if its ``definition_uri`` and ``location`` *exactly* match what's provided.
        This implies that if the location involves a module,
        it will *not* match a module that's compatible but not identical.
        """
        for candidate in reversed(list(self._state.labware_offsets_by_id.values())):
            if (
                candidate.definitionUri == definition_uri
                and candidate.location == location
            ):
                return candidate

        return None

    def get_fixed_trash_id(self) -> Optional[str]:
        """Get the identifier of labware loaded into the fixed trash location.

        Raises:
            LabwareNotLoadedError: a fixed trash was not loaded by the deck definition
                that is currently in use for the protocol run.
        """
        for labware in self._state.labware_by_id.values():
            if isinstance(
                labware.location, DeckSlotLocation
            ) and labware.location.slotName in {
                DeckSlotName.FIXED_TRASH,
                DeckSlotName.SLOT_A3,
            }:
                return labware.id
        return None

    def is_fixed_trash(self, labware_id: str) -> bool:
        """Check if labware is fixed trash."""
        return self.get_has_quirk(labware_id, "fixedTrash")

    def is_absorbance_reader_lid(self, labware_id: str) -> bool:
        """Check if labware is an absorbance reader lid."""
        return labware_validation.is_absorbance_reader_lid(
            self.get(labware_id).loadName
        )

    def raise_if_labware_inaccessible_by_pipette(self, labware_id: str) -> None:
        """Raise an error if the specified location cannot be reached via a pipette."""
        labware = self.get(labware_id)
        labware_location = labware.location
        if isinstance(labware_location, OnLabwareLocation):
            return self.raise_if_labware_inaccessible_by_pipette(
                labware_location.labwareId
            )
        elif labware.lid_id is not None:
            raise errors.LocationNotAccessibleByPipetteError(
                f"Cannot move pipette to {labware.loadName} "
                "because labware is currently covered by a lid."
            )
        elif isinstance(labware_location, AddressableAreaLocation):
            if fixture_validation.is_staging_slot(labware_location.addressableAreaName):
                raise errors.LocationNotAccessibleByPipetteError(
                    f"Cannot move pipette to {labware.loadName},"
                    f" labware is on staging slot {labware_location.addressableAreaName}"
                )
        elif (
            labware_location == OFF_DECK_LOCATION or labware_location == SYSTEM_LOCATION
        ):
            raise errors.LocationNotAccessibleByPipetteError(
                f"Cannot move pipette to {labware.loadName}, labware is off-deck."
            )

    def raise_if_labware_in_location(
        self,
        location: OnDeckLabwareLocation,
    ) -> None:
        """Raise an error if the specified location has labware in it."""
        for labware in self.get_all():
            if labware.location == location:
                raise errors.LocationIsOccupiedError(
                    f"Labware {labware.loadName} is already present at {location}."
                )

    def raise_if_labware_cannot_be_ondeck(
        self,
        location: LabwareLocation,
        labware_definition: LabwareDefinition,
    ) -> None:
        """Raise an error if the labware cannot be in the specified location."""
        if isinstance(
            location, (DeckSlotLocation, AddressableAreaLocation)
        ) and not labware_validation.validate_labware_can_be_ondeck(labware_definition):
            raise errors.LabwareCannotSitOnDeckError(
                f"{labware_definition.parameters.loadName} cannot sit in a slot by itself."
            )

    def raise_if_labware_incompatible_with_plate_reader(
        self,
        labware_definition: LabwareDefinition,
    ) -> None:
        """Raise an error if the labware is not compatible with the plate reader."""
        load_name = labware_definition.parameters.loadName
        number_of_wells = len(labware_definition.wells)
        if number_of_wells != 96:
            raise errors.LabwareMovementNotAllowedError(
                f"Cannot move '{load_name}' into plate reader because the"
                f" labware contains {number_of_wells} wells where 96 wells is expected."
            )
        elif labware_definition.dimensions.zDimension > _PLATE_READER_MAX_LABWARE_Z_MM:
            raise errors.LabwareMovementNotAllowedError(
                f"Cannot move '{load_name}' into plate reader because the"
                f" maximum allowed labware height is {_PLATE_READER_MAX_LABWARE_Z_MM}mm."
            )

    def raise_if_stacker_labware_pool_is_not_valid(
        self,
        primary_labware_definition: LabwareDefinition,
        lid_labware_definition: LabwareDefinition | None,
        adapter_labware_definition: LabwareDefinition | None,
    ) -> None:
        """Raise if the primary, lid, and adapter do not go together."""
        if lid_labware_definition:
            if not labware_validation.validate_definition_is_lid(
                lid_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {lid_labware_definition.parameters.loadName} cannot be used as a lid in the Flex Stacker."
                )
            if not labware_validation.validate_labware_can_be_stacked(
                lid_labware_definition, primary_labware_definition.parameters.loadName
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {lid_labware_definition.parameters.loadName} cannot be used as a lid for {primary_labware_definition.parameters.loadName}"
                )
        if adapter_labware_definition:
            if not labware_validation.validate_definition_is_adapter(
                adapter_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {adapter_labware_definition.parameters.loadName} cannot be used as an adapter in the Flex Stacker."
                )
            if not labware_validation.validate_labware_can_be_stacked(
                primary_labware_definition,
                adapter_labware_definition.parameters.loadName,
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {adapter_labware_definition.parameters.loadName} cannot be used as an adapter for {primary_labware_definition.parameters.loadName}"
                )

    def raise_if_labware_cannot_be_stacked(  # noqa: C901
        self, top_labware_definition: LabwareDefinition, bottom_labware_id: str
    ) -> None:
        """Raise if the specified labware definition cannot be placed on top of the bottom labware."""
        if labware_validation.validate_definition_is_adapter(top_labware_definition):
            raise errors.LabwareCannotBeStackedError(
                f"Labware {top_labware_definition.parameters.loadName} is defined as an adapter and cannot be placed"
                " on other labware."
            )
        below_labware = self.get(bottom_labware_id)
        if not labware_validation.validate_labware_can_be_stacked(
            top_labware_definition=top_labware_definition,
            below_labware_load_name=below_labware.loadName,
        ):
            raise errors.LabwareCannotBeStackedError(
                f"Labware {top_labware_definition.parameters.loadName} cannot be loaded onto labware {below_labware.loadName}"
            )
        elif (
            labware_validation.validate_definition_is_lid(top_labware_definition)
            and top_labware_definition.compatibleParentLabware is not None
            and self.get_load_name(bottom_labware_id)
            not in top_labware_definition.compatibleParentLabware
        ):
            # This parent is assumed to be compatible, unless the lid enumerates
            # all its compatible parents and this parent is missing from the list.
            raise ValueError(
                f"Labware Lid {top_labware_definition.parameters.loadName} may not be loaded on parent labware"
                f" {self.get_display_name(bottom_labware_id)}."
            )
        elif isinstance(below_labware.location, ModuleLocation):
            below_definition = self.get_definition(labware_id=below_labware.id)
            if not labware_validation.validate_definition_is_adapter(
                below_definition
            ) and not labware_validation.validate_definition_is_lid(
                top_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {top_labware_definition.parameters.loadName} cannot be loaded"
                    f" onto a labware on top of a module"
                )
        elif isinstance(below_labware.location, OnLabwareLocation):
            labware_stack = self.get_labware_stack([below_labware])
            stack_without_adapters = []
            for lw in labware_stack:
                if not labware_validation.validate_definition_is_adapter(
                    self.get_definition(lw.id)
                ) and not labware_validation.is_lid_stack(self.get_load_name(lw.id)):
                    stack_without_adapters.append(lw)
            if len(stack_without_adapters) >= self.get_labware_stacking_maximum(
                top_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {top_labware_definition.parameters.loadName} cannot be loaded to stack of more than {self.get_labware_stacking_maximum(top_labware_definition)} labware."
                )

            further_below_definition = self.get_definition(
                labware_id=below_labware.location.labwareId
            )
            if labware_validation.validate_definition_is_adapter(
                further_below_definition
            ) and not labware_validation.validate_definition_is_lid(
                top_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {top_labware_definition.parameters.loadName} cannot be loaded"
                    f" onto labware on top of adapter"
                )

    def _is_magnetic_module_uri_in_half_millimeter(self, labware_id: str) -> bool:
        """Check whether the labware uri needs to be calculated in half a millimeter."""
        uri = self.get_uri_from_definition(self.get_definition(labware_id))
        return uri in _MAGDECK_HALF_MM_LABWARE

    def get_deck_default_gripper_offsets(self) -> Optional[LabwareMovementOffsetData]:
        """Get the deck's default gripper offsets."""
        parsed_offsets = (
            self.get_deck_definition().get("gripperOffsets", {}).get("default")
        )
        return (
            LabwareMovementOffsetData(
                pickUpOffset=LabwareOffsetVector(
                    x=parsed_offsets["pickUpOffset"]["x"],
                    y=parsed_offsets["pickUpOffset"]["y"],
                    z=parsed_offsets["pickUpOffset"]["z"],
                ),
                dropOffset=LabwareOffsetVector(
                    x=parsed_offsets["dropOffset"]["x"],
                    y=parsed_offsets["dropOffset"]["y"],
                    z=parsed_offsets["dropOffset"]["z"],
                ),
            )
            if parsed_offsets
            else None
        )

    def get_absorbance_reader_lid_definition(self) -> LabwareDefinition:
        """Return the special labware definition for the plate reader lid.

        See todo comments in `create_protocol_engine().
        """
        # NOTE: This needs to stay in sync with create_protocol_engine().
        return self._state.definitions_by_uri[
            "opentrons/opentrons_flex_lid_absorbance_plate_reader_module/1"
        ]

    @overload
    def get_child_gripper_offsets(
        self,
        *,
        labware_definition: LabwareDefinition,
        slot_name: Optional[DeckSlotName],
    ) -> Optional[LabwareMovementOffsetData]:
        pass

    @overload
    def get_child_gripper_offsets(
        self, *, labware_id: str, slot_name: Optional[DeckSlotName]
    ) -> Optional[LabwareMovementOffsetData]:
        pass

    def get_child_gripper_offsets(
        self,
        *,
        labware_definition: Optional[LabwareDefinition] = None,
        labware_id: Optional[str] = None,
        slot_name: Optional[DeckSlotName],
    ) -> Optional[LabwareMovementOffsetData]:
        """Get the grip offsets that a labware says should be applied to children stacked atop it.

        Params:
            labware_id: The ID of a parent labware (atop which another labware, the child, will be stacked).
            slot_name: The ancestor slot that the parent labware is ultimately loaded into,
                       perhaps after going through a module in the middle.

        Returns:
            If `slot_name` is provided, returns the gripper offsets that the parent labware definition
            specifies just for that slot, or `None` if the labware definition doesn't have an
            exact match.

            If `slot_name` is `None`, returns the gripper offsets that the parent labware
            definition designates as "default," or `None` if it doesn't designate any as such.
        """
        if labware_id is not None:
            labware_definition = self.get_definition(labware_id)
        else:
            # Should be ensured by our @overloads.
            assert labware_definition is not None

        parsed_offsets = labware_definition.gripperOffsets
        offset_key = slot_name.id if slot_name else "default"

        if parsed_offsets is None or offset_key not in parsed_offsets:
            return None
        else:
            return LabwareMovementOffsetData(
                pickUpOffset=LabwareOffsetVector.model_construct(
                    x=parsed_offsets[offset_key].pickUpOffset.x,
                    y=parsed_offsets[offset_key].pickUpOffset.y,
                    z=parsed_offsets[offset_key].pickUpOffset.z,
                ),
                dropOffset=LabwareOffsetVector.model_construct(
                    x=parsed_offsets[offset_key].dropOffset.x,
                    y=parsed_offsets[offset_key].dropOffset.y,
                    z=parsed_offsets[offset_key].dropOffset.z,
                ),
            )

    def get_grip_force(self, labware_definition: LabwareDefinition) -> float:
        """Get the recommended grip force for gripping labware using gripper."""
        recommended_force = labware_definition.gripForce
        return (
            recommended_force if recommended_force is not None else LABWARE_GRIP_FORCE
        )

    def get_grip_height_from_labware_bottom(
        self, labware_definition: LabwareDefinition
    ) -> float:
        """Get the recommended grip height from labware bottom, if present."""
        recommended_height = labware_definition.gripHeightFromLabwareBottom
        return (
            recommended_height
            if recommended_height is not None
            else self.get_dimensions(labware_definition=labware_definition).z / 2
        )

    @staticmethod
    def _max_x_of_well(well_defn: _WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.x + (well_defn.xDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.x + (well_defn.diameter or 0) / 2
        else:
            return well_defn.x

    @staticmethod
    def _min_x_of_well(well_defn: _WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.x - (well_defn.xDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.x - (well_defn.diameter or 0) / 2
        else:
            return 0

    @staticmethod
    def _max_y_of_well(well_defn: _WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.y + (well_defn.yDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.y + (well_defn.diameter or 0) / 2
        else:
            return 0

    @staticmethod
    def _min_y_of_well(well_defn: _WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.y - (well_defn.yDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.y - (well_defn.diameter or 0) / 2
        else:
            return 0

    @staticmethod
    def _max_z_of_well(well_defn: _WellDefinition) -> float:
        return well_defn.z + well_defn.depth

    def get_well_bbox(self, labware_definition: LabwareDefinition) -> Dimensions:
        """Get the bounding box implied by the wells.

        The bounding box of the labware that is implied by the wells is that required
        to contain the bounds of the wells - the y-span from the min-y bound of the min-y
        well to the max-y bound of the max-y well, x ditto, z from labware 0 to the max-z
        well top.

        This is used for the specific purpose of finding the reasonable uncertainty bounds of
        where and how a gripper will interact with a labware.
        """
        max_x: Optional[float] = None
        min_x: Optional[float] = None
        max_y: Optional[float] = None
        min_y: Optional[float] = None
        max_z: Optional[float] = None

        for well in labware_definition.wells.values():
            well_max_x = self._max_x_of_well(well)
            well_min_x = self._min_x_of_well(well)
            well_max_y = self._max_y_of_well(well)
            well_min_y = self._min_y_of_well(well)
            well_max_z = self._max_z_of_well(well)
            if (max_x is None) or (well_max_x > max_x):
                max_x = well_max_x
            if (max_y is None) or (well_max_y > max_y):
                max_y = well_max_y
            if (min_x is None) or (well_min_x < min_x):
                min_x = well_min_x
            if (min_y is None) or (well_min_y < min_y):
                min_y = well_min_y
            if (max_z is None) or (well_max_z > max_z):
                max_z = well_max_z
        if (
            max_x is None
            or max_y is None
            or min_x is None
            or min_y is None
            or max_z is None
        ):
            return Dimensions(0, 0, 0)
        return Dimensions(max_x - min_x, max_y - min_y, max_z)



"""Load lid command request, result, and implementation models."""

from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ..errors import LabwareCannotBeStackedError, LabwareIsNotAllowedInLocationError
from ..resources import labware_validation
from ..types import (
    LoadableLabwareLocation,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
)

from .labware_handling_common import LabwareHandlingResultMixin
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import EquipmentHandler


LoadLidCommandType = Literal["loadLid"]


class LoadLidParams(BaseModel):
    """Payload required to load a lid onto a labware."""

    location: LoadableLabwareLocation = Field(
        ...,
        description="Labware the lid should be loaded onto.",
    )
    loadName: str = Field(
        ...,
        description="Name used to reference a lid labware definition.",
    )
    namespace: str = Field(
        ...,
        description="The namespace the lid labware definition belongs to.",
    )
    version: int = Field(
        ...,
        description="The lid labware definition version.",
    )


class LoadLidResult(LabwareHandlingResultMixin):
    """Result data from the execution of a LoadLabware command."""

    definition: LabwareDefinition = Field(
        ...,
        description="The full definition data for this lid labware.",
    )


class LoadLidImplementation(
    AbstractCommandImpl[LoadLidParams, SuccessData[LoadLidResult]]
):
    """Load lid command implementation."""

    def __init__(
        self, equipment: EquipmentHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._state_view = state_view

    async def execute(self, params: LoadLidParams) -> SuccessData[LoadLidResult]:
        """Load definition and calibration data necessary for a lid."""
        if not isinstance(params.location, OnLabwareLocation):
            raise LabwareIsNotAllowedInLocationError(
                "Lid Labware is only allowed to be loaded on top of a labware. Try `load_lid_stack(...)` to load lids without parent labware."
            )

        verified_location = self._state_view.geometry.ensure_location_not_occupied(
            params.location
        )
        loaded_labware = await self._equipment.load_labware(
            load_name=params.loadName,
            namespace=params.namespace,
            version=params.version,
            location=verified_location,
            labware_id=None,
        )

        if not labware_validation.validate_definition_is_lid(loaded_labware.definition):
            raise LabwareCannotBeStackedError(
                f"Labware {params.loadName} is not a Lid and cannot be loaded onto {self._state_view.labware.get_display_name(params.location.labwareId)}."
            )

        state_update = StateUpdate()

        # In the case of lids being loaded on top of other labware, set the parent labware's lid
        state_update.set_lids(
            parent_labware_ids=[params.location.labwareId],
            lid_ids=[loaded_labware.labware_id],
        )

        state_update.set_loaded_labware(
            labware_id=loaded_labware.labware_id,
            offset_id=loaded_labware.offsetId,
            definition=loaded_labware.definition,
            location=verified_location,
            display_name=None,
        )

        if isinstance(verified_location, OnLabwareLocation):
            self._state_view.labware.raise_if_labware_cannot_be_stacked(
                top_labware_definition=loaded_labware.definition,
                bottom_labware_id=verified_location.labwareId,
            )

        return SuccessData(
            public=LoadLidResult(
                labwareId=loaded_labware.labware_id,
                definition=loaded_labware.definition,
                # Note: the lid is not yet loaded and therefore won't be found as the lid id for the
                # labware onto which we're loading it, so build that part of the location sequence
                # here and then build the rest of the sequence from the parent labware
                locationSequence=[
                    OnLabwareLocationSequenceComponent(
                        labwareId=params.location.labwareId,
                        lidId=loaded_labware.labware_id,
                    )
                ]
                + self._state_view.geometry.get_location_sequence(
                    params.location.labwareId
                ),
            ),
            state_update=state_update,
        )


class LoadLid(BaseCommand[LoadLidParams, LoadLidResult, ErrorOccurrence]):
    """Load lid command resource model."""

    commandType: LoadLidCommandType = "loadLid"
    params: LoadLidParams
    result: Optional[LoadLidResult] = None

    _ImplementationCls: Type[LoadLidImplementation] = LoadLidImplementation


class LoadLidCreate(BaseCommandCreate[LoadLidParams]):
    """Load lid command creation request."""

    commandType: LoadLidCommandType = "loadLid"
    params: LoadLidParams

    _CommandCls: Type[LoadLid] = LoadLid



"""JSON file reading."""
from typing import Union

from opentrons_shared_data.protocol.models.protocol_schema_v6 import ProtocolSchemaV6
from opentrons_shared_data.protocol.models.protocol_schema_v7 import ProtocolSchemaV7
from opentrons_shared_data.protocol.models.protocol_schema_v8 import ProtocolSchemaV8
from opentrons.protocol_reader import (
    ProtocolSource,
    JsonProtocolConfig,
    ProtocolFilesInvalidError,
)


class JsonFileReader:
    """Reads and parses JSON protocol files."""

    @staticmethod
    def read(
        protocol_source: ProtocolSource,
    ) -> Union[ProtocolSchemaV6, ProtocolSchemaV7, ProtocolSchemaV8]:
        """Read and parse file into a JsonProtocol model."""
        name = protocol_source.metadata.get("name", protocol_source.main_file.name)
        if not isinstance(protocol_source.config, JsonProtocolConfig):
            raise ProtocolFilesInvalidError(
                message=f"Cannot execute {name} as a JSON protocol",
                detail={
                    "kind": "non-json-file-in-json-file-reader",
                    "metadata-name": str(protocol_source.metadata.get("name")),
                    "file-name": protocol_source.main_file.name,
                },
            )
        if protocol_source.config.schema_version == 6:
            return ProtocolSchemaV6.model_validate_json(
                protocol_source.main_file.read_bytes()
            )
        elif protocol_source.config.schema_version == 7:
            return ProtocolSchemaV7.model_validate_json(
                protocol_source.main_file.read_bytes()
            )
        elif protocol_source.config.schema_version == 8:
            return ProtocolSchemaV8.model_validate_json(
                protocol_source.main_file.read_bytes()
            )
        else:
            raise ProtocolFilesInvalidError(
                message=f"{name} is a JSON protocol v{protocol_source.config.schema_version} which this robot cannot execute",
                detail={
                    "kind": "schema-version-unknown",
                    "requested-schema-version": str(
                        protocol_source.config.schema_version
                    ),
                    "minimum-handled-schema-version": "6",
                    "maximum-handled-shcema-version": "8",
                },
            )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Dict, Union, TYPE_CHECKING, Tuple

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
)

from opentrons.protocols.advanced_control.transfers.common import (
    NoLiquidClassPropertyError,
)

from ._liquid_properties import (
    TransferProperties,
    build_transfer_properties,
)

if TYPE_CHECKING:
    from . import InstrumentContext, Labware


@dataclass(frozen=True)
class Liquid:
    """A liquid to load into a well.

    Attributes:
        name: A human-readable name for the liquid.
        description: An optional description.
        display_color: An optional display color for the liquid.

    .. versionadded:: 2.14
    """

    _id: str
    name: str
    description: Optional[str]
    display_color: Optional[str]


@dataclass
class LiquidClass:
    """A data class that contains properties of a specific class of liquids."""

    _name: str
    _display_name: str
    _by_pipette_setting: Dict[str, Dict[str, TransferProperties]]

    @classmethod
    def create(cls, liquid_class_definition: LiquidClassSchemaV1) -> "LiquidClass":
        """Liquid class factory method."""

        by_pipette_settings: Dict[str, Dict[str, TransferProperties]] = {}
        for by_pipette in liquid_class_definition.byPipette:
            tip_settings: Dict[str, TransferProperties] = {}
            for tip_type in by_pipette.byTipType:
                tip_settings[tip_type.tiprack] = build_transfer_properties(tip_type)
            by_pipette_settings[by_pipette.pipetteModel] = tip_settings

        return cls(
            _name=liquid_class_definition.liquidClassName,
            _display_name=liquid_class_definition.displayName,
            _by_pipette_setting=by_pipette_settings,
        )

    @classmethod
    def create_from(
        cls,
        name: str,
        display_name: str,
        by_pipette_setting: Dict[str, Dict[str, TransferProperties]],
    ) -> "LiquidClass":
        """Create a liquid class from the passed in args."""
        return cls(
            _name=name,
            _display_name=display_name,
            _by_pipette_setting=by_pipette_setting,
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def display_name(self) -> str:
        return self._display_name

    def update_for(
        self,
        pipette: Union[str, InstrumentContext],
        tip_rack: Union[str, Labware],
        transfer_properties: TransferProperties,
    ) -> None:
        """Update the transfer properties for the given pipette and tip combo.

        If an entry does not exist, it will be created.
        """
        pipette_name, tiprack_uri = self._get_pipette_and_tiprack_names(
            pipette, tip_rack
        )
        try:
            self._by_pipette_setting[pipette_name].update(
                {tiprack_uri: transfer_properties}
            )
        except KeyError:
            self._by_pipette_setting[pipette_name] = {tiprack_uri: transfer_properties}

    def get_for(
        self, pipette: Union[str, InstrumentContext], tip_rack: Union[str, Labware]
    ) -> TransferProperties:
        """Get liquid class transfer properties for the specified pipette and tip."""
        pipette_name, tiprack_uri = self._get_pipette_and_tiprack_names(
            pipette, tip_rack
        )

        try:
            settings_for_pipette = self._by_pipette_setting[pipette_name]
        except KeyError:
            raise NoLiquidClassPropertyError(
                f"No properties found for {pipette_name} in {self._name} liquid class"
            )
        try:
            transfer_properties = settings_for_pipette[tiprack_uri]
        except KeyError:
            raise NoLiquidClassPropertyError(
                f"No properties found for {tiprack_uri} for {pipette_name} in {self._name} liquid class"
            )
        return transfer_properties

    @staticmethod
    def _get_pipette_and_tiprack_names(
        pipette: Union[str, InstrumentContext],
        tip_rack: Union[str, Labware],
    ) -> Tuple[str, str]:
        """Return the pipette and tip rack name strings from the given pipette and tip rack."""
        from . import InstrumentContext, Labware

        if isinstance(pipette, InstrumentContext):
            pipette_name = pipette.name
        elif isinstance(pipette, str):
            pipette_name = pipette
        else:
            raise ValueError(
                f"{pipette} should either be an InstrumentContext object"
                f" or a pipette name string."
            )

        if isinstance(tip_rack, Labware):
            tiprack_uri = tip_rack.uri
        elif isinstance(tip_rack, str):
            tiprack_uri = tip_rack
        else:
            raise ValueError(
                f"{tip_rack} should either be a tiprack Labware object"
                f" or a tiprack URI string."
            )
        return pipette_name, tiprack_uri



"""Errors raised by serial connection."""

from enum import Enum
from typing import Dict, Type, Optional


class SerialException(Exception):
    """Base serial exception"""

    def __init__(self, port: str, description: str) -> None:
        super().__init__(f"{port}: {description}")
        self.port = port
        self.description = description


class NoResponse(SerialException):
    def __init__(self, port: str, command: str) -> None:
        super().__init__(port=port, description=f"No response to '{command}'")
        self.command = command


class FailedCommand(SerialException):
    def __init__(self, port: str, response: str) -> None:
        super().__init__(
            port=port, description=f"'Received error response '{response}'"
        )
        self.response = response


class AlarmResponse(FailedCommand):
    pass


class ErrorResponse(FailedCommand):
    def __init__(self, port: str, response: str, command: Optional[str] = None) -> None:
        super().__init__(port, response)
        self.command = command


class UnhandledGcode(ErrorResponse):
    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class BaseErrorCode(Enum):
    """Base class for error code enums.

    This class should be inherited to define specific sets of error codes.
    """

    @property
    def code_string(self) -> str:
        """Return the error code string."""
        code: str = self.value[0]
        return code.lower()

    @property
    def exception(self) -> Type[ErrorResponse]:
        """Return the exception class associated with this error code."""
        exc: Type[ErrorResponse] = self.value[1]
        return exc

    def raise_exception(self, port: str, response: str, command: str) -> None:
        """Raise the appropriate exception for this error code."""
        raise self.exception(port=port, response=response, command=command)

    @classmethod
    def get_error_codes(cls) -> Dict[str, "BaseErrorCode"]:
        """Get all error codes as a dictionary mapping code string to ErrorCode instance."""
        return {code.code_string: code for code in cls}


class DefaultErrorCodes(BaseErrorCode):
    """
    Default error codes that are previously handled by the SerialConnection class.
    """

    UNHANDLED_GCODE = ("ERR003", UnhandledGcode)



"""Basic liquid data state and store."""
from dataclasses import dataclass
from typing import Dict, List
from opentrons.protocol_engine.types import Liquid, LiquidId

from ._abstract_store import HasState, HandlesActions
from ..actions import Action, AddLiquidAction
from ..errors import LiquidDoesNotExistError, InvalidLiquidError


@dataclass
class LiquidState:
    """State of all loaded liquids."""

    liquids_by_id: Dict[str, Liquid]


class LiquidStore(HasState[LiquidState], HandlesActions):
    """Liquid state container."""

    _state: LiquidState

    def __init__(self) -> None:
        """Initialize a liquid store and its state."""
        self._state = LiquidState(liquids_by_id={})

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, AddLiquidAction):
            self._add_liquid(action)

    def _add_liquid(self, action: AddLiquidAction) -> None:
        """Add liquid to protocol liquids."""
        self._state.liquids_by_id[action.liquid.id] = action.liquid


class LiquidView:
    """Read-only liquid state view."""

    _state: LiquidState

    def __init__(self, state: LiquidState) -> None:
        """Initialize the computed view of liquid state.

        Arguments:
            state: Liquid state dataclass used for all calculations.
        """
        self._state = state

    def get_all(self) -> List[Liquid]:
        """Get all protocol liquids."""
        return list(self._state.liquids_by_id.values())

    def validate_liquid_id(self, liquid_id: LiquidId) -> LiquidId:
        """Check if liquid_id exists in liquids."""
        is_empty = liquid_id == "EMPTY"
        if is_empty:
            return liquid_id
        has_liquid = liquid_id in self._state.liquids_by_id
        if not has_liquid:
            raise LiquidDoesNotExistError(
                f"Supplied liquidId: {liquid_id} does not exist in the loaded liquids."
            )
        return liquid_id

    def validate_liquid_allowed(self, liquid: Liquid) -> Liquid:
        """Validate that a liquid is legal to load."""
        is_empty = liquid.id == "EMPTY"
        if is_empty:
            raise InvalidLiquidError(
                message='Protocols may not define a liquid with the special id "EMPTY".'
            )
        return liquid



""" opentrons.execute: functions and entrypoint for running protocols

This module has functions that can be imported to provide protocol
contexts for running protocols during interactive sessions like Jupyter or just
regular python shells. It also provides a console entrypoint for running a
protocol from the command line.
"""
import asyncio
import atexit
import argparse
import contextlib
import logging
import os
import sys
from typing import (
    TYPE_CHECKING,
    BinaryIO,
    Callable,
    Dict,
    List,
    Optional,
    TextIO,
    Union,
)

from opentrons_shared_data.labware.labware_definition import (
    labware_definition_type_adapter,
)
from opentrons_shared_data.robot.types import RobotType

from opentrons import protocol_api, __version__, should_use_ot3

from opentrons.legacy_commands import types as command_types

from opentrons.hardware_control import (
    API as OT2API,
    ThreadManagedHardware,
    ThreadManager,
)
from opentrons.hardware_control.types import HardwareFeatureFlags

from opentrons.protocols import parse
from opentrons.protocols.api_support.deck_type import (
    guess_from_global_config as guess_deck_type_from_global_config,
    should_load_fixed_trash,
    should_load_fixed_trash_labware_for_python_protocol,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.execution import execute as execute_apiv2
from opentrons.protocols.types import (
    ApiDeprecationError,
    Protocol,
    PythonProtocol,
)

from opentrons.protocol_api.core.engine import ENGINE_CORE_API_VERSION
from opentrons.protocol_api.protocol_context import ProtocolContext

from opentrons.protocol_engine import (
    Config,
    DeckType,
    EngineStatus,
    error_recovery_policy,
)
from opentrons.protocol_engine.create_protocol_engine import (
    create_protocol_engine_in_thread,
    create_protocol_engine,
)
from opentrons.protocol_engine.types import PostRunHardwareState

from opentrons.protocol_reader import ProtocolSource

from opentrons.protocol_runner import (
    create_protocol_runner,
    RunOrchestrator,
    LiveRunner,
)

from .util import entrypoint_util

if TYPE_CHECKING:
    from opentrons_shared_data.labware.types import (
        LabwareDefinition as LabwareDefinitionDict,
    )


_THREAD_MANAGED_HW: Optional[ThreadManagedHardware] = None
#: The background global cache that all protocol contexts created by
#: :py:meth:`get_protocol_api` will share


# When a ProtocolContext is using a ProtocolEngine to control the robot, it requires some
# additional long-lived resources besides _THREAD_MANAGED_HARDWARE. There's a background thread,
# an asyncio event loop in that thread, and some ProtocolEngine-controlled background tasks in that
# event loop.
#
# When we're executing a protocol file beginning-to-end, we can clean up those resources after it
# completes. However, when someone gets a live ProtocolContext through get_protocol_api(), we have
# no way of knowing when they're done with it. So, as a hack, we keep these resources open
# indefinitely, letting them leak.
#
# We keep this at module scope so that the contained context managers aren't garbage-collected.
# If they're garbage collected, they can close their resources prematurely.
# https://stackoverflow.com/a/69155026/497934
_LIVE_PROTOCOL_ENGINE_CONTEXTS = contextlib.ExitStack()


# See Jira RCORE-535.
_JSON_TOO_NEW_MESSAGE = (
    "Protocols created by recent versions of Protocol Designer"
    " cannot currently be executed with"
    " the opentrons_execute command-line tool"
    " or the opentrons.execute.execute() function."
    " Use the Opentrons App instead."
)


_EmitRunlogCallable = Callable[[command_types.CommandMessage], None]


def get_protocol_api(
    version: Union[str, APIVersion],
    bundled_labware: Optional[Dict[str, "LabwareDefinitionDict"]] = None,
    bundled_data: Optional[Dict[str, bytes]] = None,
    extra_labware: Optional[Dict[str, "LabwareDefinitionDict"]] = None,
    # If you add any more arguments here, make sure they're kw-only to make mistakes harder in
    # environments without type checking, like Jupyter Notebook.
    # *
) -> protocol_api.ProtocolContext:
    """
    Build and return a ``protocol_api.ProtocolContext``
    connected to the robot.

    This can be used to run protocols from interactive Python sessions
    such as Jupyter or an interpreter on the command line:

    .. code-block:: python

        >>> from opentrons.execute import get_protocol_api
        >>> protocol = get_protocol_api('2.0')
        >>> instr = protocol.load_instrument('p300_single', 'right')
        >>> instr.home()

    When this function is called, modules and instruments will be recached.

    :param version: The API version to use. This must be lower than
        ``opentrons.protocol_api.MAX_SUPPORTED_VERSION``.
        It may be specified either as a string (``'2.0'``) or
        as a ``protocols.types.APIVersion``
        (``APIVersion(2, 0)``).
    :param bundled_labware: If specified, a mapping from labware names to
        labware definitions for labware to consider in the
        protocol. Note that if you specify this, *only*
        labware in this argument will be allowed in the
        protocol. This is preparation for a beta feature
        and is best not used.
    :param bundled_data: If specified, a mapping from filenames to contents
        for data to be available in the protocol from
        :py:obj:`opentrons.protocol_api.ProtocolContext.bundled_data`.
    :param extra_labware: A mapping from labware load names to custom labware definitions.
        If this is ``None`` (the default), and this function is called on a robot,
        it will look for labware in the ``labware`` subdirectory of the Jupyter
        data directory.
    :return: The protocol context.
    """
    if isinstance(version, str):
        checked_version = parse.version_from_string(version)
    elif not isinstance(version, APIVersion):
        raise TypeError("version must be either a string or an APIVersion")
    else:
        checked_version = version

    if extra_labware is None:
        extra_labware = {
            uri: details.definition
            for uri, details in (entrypoint_util.find_jupyter_labware() or {}).items()
        }

    robot_type = _get_robot_type()
    deck_type = guess_deck_type_from_global_config()

    hardware_controller = _get_global_hardware_controller(robot_type)

    if checked_version < ENGINE_CORE_API_VERSION:
        context = _create_live_context_non_pe(
            api_version=checked_version,
            deck_type=deck_type,
            hardware_api=hardware_controller,
            bundled_labware=bundled_labware,
            bundled_data=bundled_data,
            extra_labware=extra_labware,
        )
    else:
        if bundled_labware is not None:
            # Protocol Engine has a deep assumption that standard labware definitions are always
            # implicitly loadable.
            raise NotImplementedError(
                f"The bundled_labware argument is not currently supported for Python protocols"
                f" with apiLevel {ENGINE_CORE_API_VERSION} or newer."
            )
        context = _create_live_context_pe(
            api_version=checked_version,
            robot_type=robot_type,
            deck_type=deck_type,
            hardware_api=_THREAD_MANAGED_HW,  # type: ignore[arg-type]
            bundled_data=bundled_data,
            extra_labware=extra_labware,
        )

    hardware_controller.sync.cache_instruments()
    return context


def get_arguments(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Get the argument parser for this module

    Useful if you want to use this module as a component of another CLI program
    and want to add its arguments.

    :param parser: A parser to add arguments to.
    :returns argparse.ArgumentParser: The parser with arguments added.
    """
    parser.add_argument(
        "-l",
        "--log-level",
        choices=["debug", "info", "warning", "error", "none"],
        default="warning",
        help="Specify the level filter for logs to show on the command line. "
        "The logs stored in journald or local log files are unaffected by "
        "this option and should be configured in the config file. If "
        "'none', do not show logs",
    )
    parser.add_argument(
        "-L",
        "--custom-labware-path",
        action="append",
        default=[os.getcwd()],
        help="Specify directories to search for custom labware definitions. "
        "You can specify this argument multiple times. Once you specify "
        "a directory in this way, labware definitions in that directory "
        "will become available in ProtocolContext.load_labware(). "
        "Only directories specified directly by "
        "this argument are searched, not their children. JSON files that "
        "do not define labware will be ignored with a message. "
        "The current directory (the one from which you are "
        "invoking this program) will always be included implicitly, "
        "in addition to any directories that you specify.",
    )
    parser.add_argument(
        "-D",
        "--custom-data-path",
        action="append",
        nargs="?",
        const=".",
        default=[],
        help="Specify directories to search for custom data files. "
        "You can specify this argument multiple times. Once you specify "
        "a directory in this way, files located in the specified "
        "directory will be available in ProtocolContext.bundled_data. "
        "Note that bundle execution will still only allow data files in "
        "the bundle. If you specify this without a path, it will "
        "add the current path implicitly. If you do not specify this "
        "argument at all, no data files will be added. Any file in the "
        "specified paths will be loaded into memory and included in the "
        "bundle if --bundle is passed, so be careful that any directory "
        "you specify has only the files you want. It is usually a "
        "better idea to use -d so no files are accidentally included. "
        "Also note that data files are made available as their name, not "
        "their full path, so name them uniquely.",
    )
    parser.add_argument(
        "-d",
        "--custom-data-file",
        action="append",
        default=[],
        help="Specify data files to be made available in "
        "ProtocolContext.bundled_data (and possibly bundled if --bundle "
        "is passed). Can be specified multiple times with different "
        "files. It is usually a better idea to use this than -D because "
        "there is less possibility of accidentally including something.",
    )
    parser.add_argument(
        "protocol",
        metavar="PROTOCOL",
        type=argparse.FileType("rb"),
        help="The protocol file to execute. If you pass '-', you can pipe "
        "the protocol via stdin; this could be useful if you want to use this "
        "utility as part of an automated workflow.",
    )
    return parser


def execute(
    protocol_file: Union[BinaryIO, TextIO],
    protocol_name: str,
    propagate_logs: bool = False,
    log_level: str = "warning",
    emit_runlog: Optional[_EmitRunlogCallable] = None,
    custom_labware_paths: Optional[List[str]] = None,
    custom_data_paths: Optional[List[str]] = None,
) -> None:
    """
    Run the protocol itself.

    This is a one-stop function to run a protocol, whether python or json,
    no matter the api version, from external (i.e. not bound up in other
    internal server infrastructure) sources.

    To run an opentrons protocol from other places, pass in a file like
    object as protocol_file; this function either returns (if the run has no
    problems) or raises an exception.

    To call from the command line use either the autogenerated entrypoint
    ``opentrons_execute`` or ``python -m opentrons.execute``.

    :param protocol_file: The protocol file to execute
    :param protocol_name: The name of the protocol file. This is required
        internally, but it may not be a thing we can get
        from the ``protocol_file`` argument.
    :param propagate_logs: Whether this function should allow logs from the
        Opentrons stack to propagate up to the root handler.
        This can be useful if you're integrating this
        function in a larger application, but most logs that
        occur during protocol simulation are best associated
        with the actions in the protocol that cause them.
        Default: ``False``
    :param log_level: The level of logs to emit on the command line:
        ``"debug"``, ``"info"``, ``"warning"``, or ``"error"``.
        Defaults to ``"warning"``.
    :param emit_runlog: A callback for printing the run log. If specified, this
        will be called whenever a command adds an entry to the
        run log, which can be used for display and progress
        estimation. If specified, the callback should take a
        single argument (the name doesn't matter) which will
        be a dictionary:

        .. code-block:: python

          {
            'name': command_name,
            'payload': {
              'text': string_command_text,
              # The rest of this struct is
              # command-dependent; see
              # opentrons.legacy_commands.commands.
             }
          }

        .. note::
          In older software versions, ``payload["text"]`` was a
          `format string <https://docs.python.org/3/library/string.html#formatstrings>`_.
          To get human-readable text, you had to do ``payload["text"].format(**payload)``.
          Don't do that anymore. If ``payload["text"]`` happens to contain any
          ``{`` or ``}`` characters, it can confuse ``.format()`` and cause it to raise a
          ``KeyError``.

    :param custom_labware_paths: A list of directories to search for custom labware.
        Loads valid labware from these paths and makes them available
        to the protocol context. If this is ``None`` (the default), and
        this function is called on a robot, it will look in the ``labware``
        subdirectory of the Jupyter data directory.
    :param custom_data_paths: A list of directories or files to load custom
        data files from. Ignored if the apiv2 feature
        flag if not set. Entries may be either files or
        directories. Specified files and the
        non-recursive contents of specified directories
        are presented by the protocol context in
        ``ProtocolContext.bundled_data``.
    """
    stack_logger = logging.getLogger("opentrons")
    stack_logger.propagate = propagate_logs
    stack_logger.setLevel(getattr(logging, log_level.upper(), logging.WARNING))
    # TODO(mm, 2023-11-20): We should restore the original log settings when we're done.

    # TODO(mm, 2023-10-02): Switch this truthy check to `is not None`
    # to match documented behavior.
    # See notes in https://github.com/Opentrons/opentrons/pull/13107
    if custom_labware_paths:
        extra_labware = entrypoint_util.labware_from_paths(custom_labware_paths)
    else:
        extra_labware = entrypoint_util.find_jupyter_labware() or {}

    if custom_data_paths:
        extra_data = entrypoint_util.datafiles_from_paths(custom_data_paths)
    else:
        extra_data = {}

    contents = protocol_file.read()
    try:
        protocol = parse.parse(
            contents,
            protocol_name,
            extra_labware={
                uri: details.definition for uri, details in extra_labware.items()
            },
            extra_data=extra_data,
        )
    except parse.JSONSchemaVersionTooNewError as e:
        # opentrons.protocols.parse() doesn't support new JSON protocols.
        # The code to do that should be moved from opentrons.protocol_reader.
        # See https://opentrons.atlassian.net/browse/PLAT-94.
        raise NotImplementedError(_JSON_TOO_NEW_MESSAGE) from e

    if protocol.api_level < APIVersion(2, 0):
        raise ApiDeprecationError(version=protocol.api_level)

    # Guard against trying to run protocols for the wrong robot type.
    # This matches what robot-server does.
    # FIXME: This exposes the internal strings "OT-2 Standard" and "OT-3 Standard".
    # https://opentrons.atlassian.net/browse/RSS-370
    if protocol.robot_type != _get_robot_type():
        raise RuntimeError(
            f'This robot is of type "{_get_robot_type()}",'
            f' so it can\'t execute protocols for robot type "{protocol.robot_type}"'
        )

    if protocol.api_level < ENGINE_CORE_API_VERSION:
        _run_file_non_pe(
            protocol=protocol,
            emit_runlog=emit_runlog,
        )
    else:
        # TODO(mm, 2023-07-06): Once these NotImplementedErrors are resolved, consider removing
        # the enclosing if-else block and running everything through _run_file_pe() for simplicity.
        if custom_data_paths:
            raise NotImplementedError(
                f"The custom_data_paths argument is not currently supported for Python protocols"
                f" with apiLevel {ENGINE_CORE_API_VERSION} or newer."
            )
        protocol_file.seek(0)
        _run_file_pe(
            protocol=protocol,
            hardware_api=_get_global_hardware_controller(_get_robot_type()),
            emit_runlog=emit_runlog,
        )


def make_runlog_cb() -> Callable[[command_types.CommandMessage], None]:
    level = 0
    last_dollar = None

    def _print_runlog(command: command_types.CommandMessage) -> None:
        nonlocal level
        nonlocal last_dollar

        if last_dollar == command["$"]:
            if command["$"] == "before":
                level += 1
            else:
                level -= 1
        last_dollar = command["$"]
        if command["$"] == "before":
            print(" ".join(["\t" * level, command["payload"].get("text", "")]))

    return _print_runlog


def main() -> int:
    """Handler for command line invocation to run a protocol.

    :param argv: The arguments the program was invoked with; this is usually
        :py:obj:`sys.argv` but if you want to override that you can.
    :returns int: A success or failure value suitable for use as a shell
        return code passed to :py:obj:`sys.exit` (0 means success,
        anything else is a kind of failure).
    """
    parser = argparse.ArgumentParser(
        prog="opentrons_execute", description="Run an OT-2 protocol"
    )
    parser = get_arguments(parser)
    # don't want to add this in get_arguments because if somebody upstream is
    # using that parser they probably want their own version
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument(
        "-n",
        "--no-print-runlog",
        action="store_true",
        help="Do not print the commands as they are executed",
    )
    args = parser.parse_args()
    printer = None if args.no_print_runlog else make_runlog_cb()
    if args.log_level != "none":
        stack_logger = logging.getLogger("opentrons")
        stack_logger.addHandler(logging.StreamHandler(sys.stdout))
        log_level = args.log_level
    else:
        # TODO(mm, 2023-07-13): This default logging prints error information redundantly
        # when executing via Protocol Engine, because Protocol Engine logs when commands fail.
        log_level = "warning"

    try:
        execute(
            protocol_file=args.protocol,
            protocol_name=args.protocol.name,
            custom_labware_paths=args.custom_labware_path,
            custom_data_paths=(args.custom_data_path + args.custom_data_file),
            log_level=log_level,
            emit_runlog=printer,
        )
        return 0
    except entrypoint_util.ProtocolEngineExecuteError as error:
        # This exception is a wrapper that's meaningless to the CLI user.
        # Take the actual protocol problem out of it and just print that.
        print(error.to_stderr_string(), file=sys.stderr)
        return 1
    # execute() might raise other exceptions, but we don't have a nice way to print those.
    # Just let Python show a traceback.


def _create_live_context_non_pe(
    api_version: APIVersion,
    hardware_api: ThreadManagedHardware,
    deck_type: str,
    extra_labware: Optional[Dict[str, "LabwareDefinitionDict"]],
    bundled_labware: Optional[Dict[str, "LabwareDefinitionDict"]],
    bundled_data: Optional[Dict[str, bytes]],
) -> ProtocolContext:
    """Return a live ProtocolContext.

    This controls the robot through the older infrastructure, instead of through Protocol Engine.
    """
    assert api_version < ENGINE_CORE_API_VERSION
    return protocol_api.create_protocol_context(
        api_version=api_version,
        deck_type=deck_type,
        hardware_api=hardware_api,
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
        extra_labware=extra_labware,
    )


def _create_live_context_pe(
    api_version: APIVersion,
    hardware_api: ThreadManagedHardware,
    robot_type: RobotType,
    deck_type: str,
    extra_labware: Dict[str, "LabwareDefinitionDict"],
    bundled_data: Optional[Dict[str, bytes]],
) -> ProtocolContext:
    """Return a live ProtocolContext that controls the robot through ProtocolEngine."""
    assert api_version >= ENGINE_CORE_API_VERSION

    global _LIVE_PROTOCOL_ENGINE_CONTEXTS
    hardware_api_wrapped = hardware_api.wrapped()
    pe, loop = _LIVE_PROTOCOL_ENGINE_CONTEXTS.enter_context(
        create_protocol_engine_in_thread(
            hardware_api=hardware_api_wrapped,
            config=_get_protocol_engine_config(),
            deck_configuration=entrypoint_util.get_deck_configuration(),
            file_provider=None,
            error_recovery_policy=error_recovery_policy.never_recover,
            drop_tips_after_run=False,
            post_run_hardware_state=PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
            load_fixed_trash=should_load_fixed_trash_labware_for_python_protocol(
                api_version
            ),
        )
    )

    # `async def` so we can use loop.run_coroutine_threadsafe() to wait for its completion.
    # Non-async would use call_soon_threadsafe(), which makes the waiting harder.
    async def add_all_extra_labware() -> None:
        for labware_definition_dict in extra_labware.values():
            labware_definition = labware_definition_type_adapter.validate_python(
                labware_definition_dict
            )
            pe.add_labware_definition(labware_definition)

    # Add extra_labware to ProtocolEngine, being careful not to modify ProtocolEngine from this
    # thread. See concurrency notes in ProtocolEngine docstring.
    future = asyncio.run_coroutine_threadsafe(add_all_extra_labware(), loop)
    future.result()

    return protocol_api.create_protocol_context(
        api_version=api_version,
        hardware_api=hardware_api,
        deck_type=deck_type,
        protocol_engine=pe,
        protocol_engine_loop=loop,
        bundled_data=bundled_data,
    )


def _run_file_non_pe(
    protocol: Protocol,
    emit_runlog: Optional[_EmitRunlogCallable],
) -> None:
    """Run a protocol file without Protocol Engine, with the older infrastructure instead."""
    if isinstance(protocol, PythonProtocol):
        extra_labware = protocol.extra_labware
        bundled_labware = protocol.bundled_labware
        bundled_data = protocol.bundled_data
    else:
        # JSON protocols do have "bundled labware" embedded in them, but those aren't represented in
        # the parsed Protocol object and we don't need to create the ProtocolContext with them.
        # execute_apiv2.run_protocol() will pull them out of the JSON and load them into the
        # ProtocolContext.
        extra_labware = None
        bundled_labware = None
        bundled_data = None

    context = _create_live_context_non_pe(
        api_version=protocol.api_level,
        hardware_api=_get_global_hardware_controller(_get_robot_type()),
        deck_type=guess_deck_type_from_global_config(),
        extra_labware=extra_labware,
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
    )

    if emit_runlog:
        context.broker.subscribe(command_types.COMMAND, emit_runlog)

    context.home()
    try:
        execute_apiv2.run_protocol(
            protocol, context, run_time_parameters_with_overrides=None
        )
    finally:
        context.cleanup()


def _run_file_pe(
    protocol: Protocol,
    hardware_api: ThreadManagedHardware,
    emit_runlog: Optional[_EmitRunlogCallable],
) -> None:
    """Run a protocol file with Protocol Engine."""

    async def run(protocol_source: ProtocolSource) -> None:
        # TODO (spp, 2024-03-18): use run-time param overrides once enabled for cli protocol execution
        hardware_api_wrapped = hardware_api.wrapped()
        protocol_engine = await create_protocol_engine(
            hardware_api=hardware_api_wrapped,
            config=_get_protocol_engine_config(),
            error_recovery_policy=error_recovery_policy.never_recover,
            load_fixed_trash=should_load_fixed_trash(protocol_source.config),
        )

        protocol_runner = create_protocol_runner(
            protocol_config=protocol_source.config,
            protocol_engine=protocol_engine,
            hardware_api=hardware_api_wrapped,
        )

        orchestrator = RunOrchestrator(
            hardware_api=hardware_api_wrapped,
            protocol_engine=protocol_engine,
            json_or_python_protocol_runner=protocol_runner,
            fixit_runner=LiveRunner(
                protocol_engine=protocol_engine, hardware_api=hardware_api_wrapped
            ),
            setup_runner=LiveRunner(
                protocol_engine=protocol_engine, hardware_api=hardware_api_wrapped
            ),
            protocol_live_runner=LiveRunner(
                protocol_engine=protocol_engine, hardware_api=hardware_api_wrapped
            ),
        )

        unsubscribe = protocol_runner.broker.subscribe(
            "command", lambda event: emit_runlog(event) if emit_runlog else None
        )
        try:
            # TODO(mm, 2023-06-30): This will home and drop tips at the end, which is not how
            # things have historically behaved with PAPIv2.13 and older or JSONv5 and older.
            result = await orchestrator.run(
                deck_configuration=entrypoint_util.get_deck_configuration(),
                protocol_source=protocol_source,
            )
        finally:
            unsubscribe()

        if result.state_summary.status != EngineStatus.SUCCEEDED:
            raise entrypoint_util.ProtocolEngineExecuteError(
                result.state_summary.errors
            )

    with entrypoint_util.adapt_protocol_source(protocol) as protocol_source:
        asyncio.run(run(protocol_source))


def _get_robot_type() -> RobotType:
    """Return what kind of robot we're currently running on."""
    return "OT-3 Standard" if should_use_ot3() else "OT-2 Standard"


def _get_protocol_engine_config() -> Config:
    """Return a Protocol Engine config to execute protocols on this device."""
    return Config(
        robot_type=_get_robot_type(),
        deck_type=DeckType(guess_deck_type_from_global_config()),
        # We deliberately omit ignore_pause=True because, in the current implementation of
        # opentrons.protocol_api.core.engine, that would incorrectly make
        # ProtocolContext.is_simulating() return True.
    )


def _get_global_hardware_controller(robot_type: RobotType) -> ThreadManagedHardware:
    # Build a hardware controller in a worker thread, which is necessary
    # because ipython runs its notebook in asyncio but the notebook
    # is at script/repl scope not function scope and is synchronous so
    # you can't control the loop from inside. If we update to
    # IPython 7 we can avoid this, but for now we can't
    global _THREAD_MANAGED_HW
    if not _THREAD_MANAGED_HW:
        if robot_type == "OT-3 Standard":
            # Conditional import because this isn't installed on OT-2s.
            from opentrons.hardware_control.ot3api import OT3API

            _THREAD_MANAGED_HW = ThreadManager(
                OT3API.build_hardware_controller,
                feature_flags=HardwareFeatureFlags.build_from_ff(),
            )
        else:
            _THREAD_MANAGED_HW = ThreadManager(
                OT2API.build_hardware_controller,
                feature_flags=HardwareFeatureFlags.build_from_ff(),
            )

    return _THREAD_MANAGED_HW


@atexit.register
def _clear_cached_hardware_controller() -> None:
    global _THREAD_MANAGED_HW
    if _THREAD_MANAGED_HW:
        _THREAD_MANAGED_HW.clean_up()
        _THREAD_MANAGED_HW = None


# This atexit registration must come after _clear_cached_hardware_controller()
# to ensure we tear things down in order from highest level to lowest level.
@atexit.register
def _clear_live_protocol_engine_contexts() -> None:
    global _LIVE_PROTOCOL_ENGINE_CONTEXTS
    _LIVE_PROTOCOL_ENGINE_CONTEXTS.close()


if __name__ == "__main__":
    sys.exit(main())



"""Utilities for doing coverage math on wells."""

from typing import Iterator
from opentrons_shared_data.errors.exceptions import (
    InvalidStoredData,
    InvalidProtocolData,
)

from opentrons.hardware_control.nozzle_manager import NozzleMap


def wells_covered_by_pipette_configuration(
    nozzle_map: NozzleMap,
    target_well: str,
    labware_wells_by_column: list[list[str]],
) -> Iterator[str]:
    """Compute the wells covered by a pipette nozzle configuration."""
    if len(labware_wells_by_column) >= 12 and len(labware_wells_by_column[0]) >= 8:
        yield from wells_covered_dense(
            nozzle_map,
            target_well,
            labware_wells_by_column,
        )
    elif len(labware_wells_by_column) < 12 and len(labware_wells_by_column[0]) < 8:
        yield from wells_covered_sparse(
            nozzle_map, target_well, labware_wells_by_column
        )
    else:
        raise InvalidStoredData(
            "Labware of non-SBS and non-reservoir format cannot be handled"
        )


def row_col_ordinals_from_column_major_map(
    target_well: str, column_major_wells: list[list[str]]
) -> tuple[int, int]:
    """Turn a well name into the index of its row and column (in that order) within the labware."""
    for column_index, column in enumerate(column_major_wells):
        if target_well in column:
            return column.index(target_well), column_index
    raise InvalidStoredData(f"Well name {target_well} is not present in labware")


def wells_covered_dense(  # noqa: C901
    nozzle_map: NozzleMap, target_well: str, target_wells_by_column: list[list[str]]
) -> Iterator[str]:
    """Get the list of wells covered by a nozzle map on an SBS format labware with a specified multiplier of 96 into the number of wells.

    This will handle the offsetting of the nozzle map into higher-density well plates. For instance, a full column config target at A1 of a
    96 plate would cover wells A1, B1, C1, D1, E1, F1, G1, H1, and use downsample_factor 1.0 (96*1 = 96). A full column config target on a
    384 plate would cover wells A1, C1, E1, G1, I1, K1, M1, O1 and use downsample_factor 4.0 (96*4 = 384), while a full column config
    targeting B1 would cover wells B1, D1, F1, H1, J1, L1, N1, P1 - still using downsample_factor 4.0, with the offset gathered from the
    target well.

    The function may also handle sub-96 regular labware with fractional downsample factors, but that's physically improbable and it's not
    tested. If you have a regular labware with fewer than 96 wells that is still regularly-spaced and has little enough space between well
    walls that it's reasonable to use with multiple channels, you probably want wells_covered_trough.
    """
    target_row_index, target_column_index = row_col_ordinals_from_column_major_map(
        target_well, target_wells_by_column
    )
    column_downsample = len(target_wells_by_column) // 12
    row_downsample = len(target_wells_by_column[0]) // 8
    if column_downsample < 1 or row_downsample < 1:
        raise InvalidStoredData(
            "This labware cannot be used with wells_covered_dense() because it is less dense than an SBS 96 standard"
        )

    for nozzle_column in range(len(nozzle_map.columns)):
        target_column_offset = nozzle_column * column_downsample
        for nozzle_row in range(len(nozzle_map.rows)):
            target_row_offset = nozzle_row * row_downsample
            if nozzle_map.starting_nozzle == "A1":
                if (
                    target_column_index + target_column_offset
                    < len(target_wells_by_column)
                ) and (
                    target_row_index + target_row_offset
                    < len(target_wells_by_column[target_column_index])
                ):
                    yield target_wells_by_column[
                        target_column_index + target_column_offset
                    ][target_row_index + target_row_offset]
            elif nozzle_map.starting_nozzle == "A12":
                if (target_column_index - target_column_offset >= 0) and (
                    target_row_index + target_row_offset
                    < len(target_wells_by_column[target_column_index])
                ):
                    yield target_wells_by_column[
                        target_column_index - target_column_offset
                    ][target_row_index + target_row_offset]
            elif nozzle_map.starting_nozzle == "H1":
                if (
                    target_column_index + target_column_offset
                    < len(target_wells_by_column)
                ) and (target_row_index - target_row_offset >= 0):
                    yield target_wells_by_column[
                        target_column_index + target_column_offset
                    ][target_row_index - target_row_offset]
            elif nozzle_map.starting_nozzle == "H12":
                if (target_column_index - target_column_offset >= 0) and (
                    target_row_index - target_row_offset >= 0
                ):
                    yield target_wells_by_column[
                        target_column_index - target_column_offset
                    ][target_row_index - target_row_offset]
            else:
                raise InvalidProtocolData(
                    f"A pipette nozzle configuration may not having a starting nozzle of {nozzle_map.starting_nozzle}"
                )


def wells_covered_sparse(  # noqa: C901
    nozzle_map: NozzleMap, target_well: str, target_wells_by_column: list[list[str]]
) -> Iterator[str]:
    """Get the list of wells covered by a nozzle map on a column-oriented reservoir.

    This function handles reservoirs whose wells span multiple rows and columns - the most common case is something like a
    12-well reservoir, whose wells are the height of an SBS column and the width of an SBS row, or a 1-well reservoir whose well
    is the size of an SBS active area.
    """
    target_row_index, target_column_index = row_col_ordinals_from_column_major_map(
        target_well, target_wells_by_column
    )
    column_upsample = 12 // len(target_wells_by_column)
    row_upsample = 8 // len(target_wells_by_column[0])
    if column_upsample < 1 or row_upsample < 1:
        raise InvalidStoredData(
            "This labware cannot be used with wells_covered_sparse() because it is more dense than an SBS 96 standard."
        )
    for nozzle_column in range(max(1, len(nozzle_map.columns) // column_upsample)):
        for nozzle_row in range(max(1, len(nozzle_map.rows) // row_upsample)):
            if nozzle_map.starting_nozzle == "A1":
                if (
                    target_column_index + nozzle_column < len(target_wells_by_column)
                ) and (
                    target_row_index + nozzle_row
                    < len(target_wells_by_column[target_column_index])
                ):
                    yield target_wells_by_column[target_column_index + nozzle_column][
                        target_row_index + nozzle_row
                    ]
            elif nozzle_map.starting_nozzle == "A12":
                if (target_column_index - nozzle_column >= 0) and (
                    target_row_index + nozzle_row
                    < len(target_wells_by_column[target_column_index])
                ):
                    yield target_wells_by_column[target_column_index - nozzle_column][
                        target_row_index + nozzle_row
                    ]
            elif nozzle_map.starting_nozzle == "H1":
                if (
                    target_column_index + nozzle_column
                    < len(target_wells_by_column[target_column_index])
                ) and (target_row_index - nozzle_row >= 0):
                    yield target_wells_by_column[target_column_index + nozzle_column][
                        target_row_index - nozzle_row
                    ]
            elif nozzle_map.starting_nozzle == "H12":
                if (target_column_index - nozzle_column >= 0) and (
                    target_row_index - nozzle_row >= 0
                ):
                    yield target_wells_by_column[target_column_index - nozzle_column][
                        target_row_index - nozzle_row
                    ]
            else:
                raise InvalidProtocolData(
                    f"A pipette nozzle configuration may not having a starting nozzle of {nozzle_map.starting_nozzle}"
                )


def nozzles_per_well(
    nozzle_map: NozzleMap, target_well: str, target_wells_by_column: list[list[str]]
) -> int:
    """Get the number of nozzles that will interact with each well in the labware.

    For instance, if this is an SBS 96 or more dense, there is always 1 nozzle per well
    that is interacted with (and some wells may not be interacted with at all). If this is
    a 12-column reservoir, then all active nozzles in each column of the configuration will
    interact with each well; so an 8-channel full config would have 8 nozzles per well,
    and a 96 channel with a rectangle config from A1 to D12 would have 4 nozzles per well.
    """
    _, target_column_index = row_col_ordinals_from_column_major_map(
        target_well, target_wells_by_column
    )
    # labware as or more dense than a 96 plate will only ever have 1 nozzle per well (and some wells won't be touched)
    if len(target_wells_by_column) >= len(nozzle_map.columns) and len(
        target_wells_by_column[target_column_index]
    ) >= len(nozzle_map.rows):
        return 1
    return max(1, len(nozzle_map.columns) // len(target_wells_by_column)) * max(
        1, len(nozzle_map.rows) // len(target_wells_by_column[target_column_index])
    )



"""The Proxy class module."""
from __future__ import annotations
import asyncio
import logging
import uuid
import socket
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

from opentrons.hardware_control.emulation.settings import ProxySettings

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class Connection:
    """Attributes of a client connected on the server port (module emulator)."""

    identifier: str
    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter


class ProxyListener(ABC):
    """Interface defining an object needing to know when a server (module emulator)
    connected/disconnected."""

    @abstractmethod
    def on_server_connected(
        self, server_type: str, client_uri: str, identifier: str
    ) -> None:
        """Called when a new server connects."""
        ...

    @abstractmethod
    def on_server_disconnected(self, identifier: str) -> None:
        """Called when a server disconnects."""
        ...


class Proxy:
    """A class that has servers (emulators) connect on one port and clients
    (drivers) on another. A server connection will be added to a collection. A
    client connection will check for a server connection and if available will
    have its write stream attached to the servers read stream and vice versa."""

    def __init__(
        self, name: str, listener: ProxyListener, settings: ProxySettings
    ) -> None:
        """Constructor.

        Args:
            name: Proxy name.
            listener: Connection even listener.
            settings:  The proxy settings.
        """
        self._name = name
        self._settings = settings
        self._event_listener = listener
        self._cons: List[Connection] = []

    @property
    def name(self) -> str:
        """Return the name of the proxy."""
        return self._name

    async def run(self) -> None:
        """Run the server."""
        await asyncio.gather(
            self._listen_server_connections(),
            self._listen_client_connections(),
        )

    async def _listen_server_connections(self) -> None:
        """Run the server listener."""
        log.info(
            f"starting {self._name} server connection server at "
            f"{self._settings.host}:{self._settings.emulator_port}"
        )
        server = await asyncio.start_server(
            self._handle_server_connection,
            self._settings.host,
            self._settings.emulator_port,
        )
        await server.serve_forever()

    async def _listen_client_connections(self) -> None:
        """Run the client listener."""
        log.info(
            f"starting {self._name} client connection server at "
            f"{self._settings.host}:{self._settings.driver_port}"
        )
        server = await asyncio.start_server(
            self._handle_client_connection,
            self._settings.host,
            self._settings.driver_port,
        )
        await server.serve_forever()

    async def _handle_server_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Handle a server connection.

        A new connection will be added to the our connection collection.

        Args:
            reader: Reader
            writer: Writer

        Returns:
            None
        """
        log.info(f"{self._name} emulator connected.")
        connection = Connection(
            identifier=str(uuid.uuid1()), reader=reader, writer=writer
        )

        log.info(f"Using Local Host: {self._settings.use_local_host}")

        client_host = (
            "127.0.0.1" if self._settings.use_local_host else socket.gethostname()
        )

        self._cons.append(connection)
        self._event_listener.on_server_connected(
            server_type=self._name,
            client_uri=f"socket://{client_host}:{self._settings.driver_port}",
            identifier=connection.identifier,
        )

    async def _handle_client_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Handle a client connection.

        Attempt to match the client steam with an available server stream.

        Args:
            reader: Reader
            writer: Writer

        Returns:
            None
        """
        try:
            while True:
                # Pop an emulator connection.
                connection = self._cons.pop(0)
                if not connection.reader.at_eof():
                    break
                else:
                    log.info(f"{self._name} server connection terminated")
                    self._event_listener.on_server_disconnected(connection.identifier)
        except IndexError:
            log.info(f"{self._name} no emulator connected.")
            writer.close()
            return

        log.info(
            f"{self._name} "
            f"client at {writer.transport.get_extra_info('socket')}"
            f" connected to {connection.writer.transport.get_extra_info('socket')}."
        )

        await self._handle_proxy(
            driver=Connection(
                reader=reader, writer=writer, identifier=connection.identifier
            ),
            server=connection,
        )

        # Return the emulator connection to the pool.
        if not connection.reader.at_eof():
            log.info(f"{self._name} returning connection to pool")
            self._cons.append(connection)
        else:
            log.info(f"{self._name} server connection terminated")
            self._event_listener.on_server_disconnected(connection.identifier)

    async def _handle_proxy(self, driver: Connection, server: Connection) -> None:
        """Connect the driver to the emulator.

        Args:
            driver: Driver connection
            server: Emulator connection

        Returns:
            None
        """
        loop = asyncio.get_event_loop()
        read_from_client_task = loop.create_task(
            self._data_router(driver, server, False)
        )
        read_from_server_task = loop.create_task(
            self._data_router(server, driver, True)
        )
        await read_from_client_task
        read_from_server_task.cancel()
        try:
            await read_from_server_task
        except asyncio.CancelledError:
            log.exception("Server task cancelled")
            pass

    @staticmethod
    async def _data_router(
        in_connection: Connection,
        out_connection: Connection,
        close_other_on_disconnect: bool,
    ) -> None:
        """Route date from in to out.

        Args:
            in_connection: connection to read from.
            out_connection: connection to write to
            close_other_on_disconnect: whether the other connection should
                be closed if the in_connection is closed.
                A driver disconnect should close connection with emulator, while
                an emulator disconnect should close the attached driver.

        Returns:
            None
        """
        while True:
            try:
                d = await in_connection.reader.read(1)
                out_connection.writer.write(d)

                if not d:
                    log.info(
                        f"{in_connection.writer.transport.get_extra_info('socket')} "
                        f"disconnected."
                    )
                    break
            except ConnectionError:
                log.exception("connection error in data router")
                break
        if close_other_on_disconnect:
            out_connection.writer.close()



"""Comment command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional, Type
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence

CommentCommandType = Literal["comment"]


class CommentParams(BaseModel):
    """Payload required to annotate execution with a comment."""

    message: str = Field(
        ...,
        description="A user-facing message",
    )


class CommentResult(BaseModel):
    """Result data from the execution of a Comment command."""


class CommentImplementation(
    AbstractCommandImpl[CommentParams, SuccessData[CommentResult]]
):
    """Comment command implementation."""

    def __init__(self, **kwargs: object) -> None:
        pass

    async def execute(self, params: CommentParams) -> SuccessData[CommentResult]:
        """No operation taken other than capturing message in command."""
        return SuccessData(
            public=CommentResult(),
        )


class Comment(BaseCommand[CommentParams, CommentResult, ErrorOccurrence]):
    """Comment command model."""

    commandType: CommentCommandType = "comment"
    params: CommentParams
    result: Optional[CommentResult] = None

    _ImplementationCls: Type[CommentImplementation] = CommentImplementation


class CommentCreate(BaseCommandCreate[CommentParams]):
    """Comment command request model."""

    commandType: CommentCommandType = "comment"
    params: CommentParams

    _CommandCls: Type[Comment] = Comment



from __future__ import annotations
import enum
from math import sqrt, isclose
from typing import (
    TYPE_CHECKING,
    Any,
    NamedTuple,
    Iterator,
    Union,
    List,
    Optional,
    Protocol,
    Dict,
)

from opentrons_shared_data.robot.types import RobotType

from .protocols.api_support.labware_like import LabwareLike

if TYPE_CHECKING:
    from .protocol_api.labware import Labware, Well
    from .protocol_api.core.legacy.module_geometry import ModuleGeometry
    from .protocol_api.module_contexts import ModuleContext
    from .protocol_api._types import OffDeckType


class PipetteNotAttachedError(KeyError):
    """An error raised if a pipette is accessed that is not attached"""

    pass


class Point(NamedTuple):
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __add__(self, other: Any) -> Point:
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: Any) -> Point:
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: Union[int, float]) -> Point:  # type: ignore[override]
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Point(self.x * other, self.y * other, self.z * other)

    def __rmul__(self, other: Union[int, float]) -> Point:  # type: ignore[override]
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Point(self.x * other, self.y * other, self.z * other)

    def __abs__(self) -> Point:
        return Point(abs(self.x), abs(self.y), abs(self.z))

    def __str__(self) -> str:
        return "({}, {}, {})".format(self.x, self.y, self.z)

    def magnitude_to(self, other: Any) -> float:
        if not isinstance(other, Point):
            return NotImplemented
        x_diff = self.x - other.x
        y_diff = self.y - other.y
        z_diff = self.z - other.z
        return sqrt(x_diff**2 + y_diff**2 + z_diff**2)

    def elementwise_isclose(
        self, other: Point, *, rel_tol: float = 1e-05, abs_tol: float = 1e-08
    ) -> bool:
        pairs = ((self.x, other.x), (self.y, other.y), (self.z, other.z))
        return all(isclose(s, o, rel_tol=rel_tol, abs_tol=abs_tol) for s, o in pairs)


LocationLabware = Union[
    "Labware",
    "Well",
    str,
    "ModuleGeometry",
    LabwareLike,
    None,
    "OffDeckType",
    "ModuleContext",
]


class MeniscusTrackingTarget(enum.Enum):
    START = "start"
    END = "end"
    DYNAMIC = "dynamic"

    def __str__(self) -> str:
        return self.name


class Location:
    """Location(point: Point, labware: Union["Labware", "Well", str, "ModuleGeometry", LabwareLike, None, "ModuleContext"])

    A location to target as a motion.

    The location contains a :py:class:`.Point` (in
    :ref:`protocol-api-deck-coords`) and possibly an associated
    :py:class:`.Labware` or :py:class:`.Well` instance.

    It should rarely be constructed directly by the user; rather, it is the
    return type of most :py:class:`.Well` accessors like :py:meth:`.Well.top`
    and is passed directly into a method like ``InstrumentContext.aspirate()``.

    .. warning::
       The ``.labware`` attribute of this class is used by the protocol
       API internals to, among other things, determine safe heights to retract
       the instruments to when moving between locations. If constructing an
       instance of this class manually, be sure to either specify ``None`` as the
       labware (so the robot does its worst case retraction) or specify the
       correct labware for the ``.point`` attribute.


    .. warning::
       The ``==`` operation compares both the position and associated labware.
       If you only need to compare locations, compare the ``.point``
       of each item.
    """

    def __init__(
        self,
        point: Point,
        labware: Union[
            "Labware",
            "Well",
            str,
            "ModuleGeometry",
            LabwareLike,
            None,
            "ModuleContext",
        ],
        *,
        _meniscus_tracking: Optional[MeniscusTrackingTarget] = None,
    ):
        self._point = point
        self._given_labware = labware
        self._labware = LabwareLike(labware)
        self._meniscus_tracking = _meniscus_tracking

    # todo(mm, 2021-10-01): Figure out how to get .point and .labware to show up
    # in the rendered docs, and then update the class docstring to use cross-references.

    @property
    def point(self) -> Point:
        return self._point

    @property
    def labware(self) -> LabwareLike:
        return self._labware

    @property
    def meniscus_tracking(self) -> Optional[MeniscusTrackingTarget]:
        return self._meniscus_tracking

    def __iter__(self) -> Iterator[Union[Point, LabwareLike]]:
        """Iterable interface to support unpacking. Like a tuple.

        .. note::
           While type annotations cannot properly support this, it will work in practice:

           point, labware = location
           some_function_taking_both(*location)
        """
        return iter((self._point, self._labware))

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Location)
            and other._point == self._point
            and other._labware == self._labware
            and other._meniscus_tracking == self._meniscus_tracking
        )

    def move(self, point: Point) -> "Location":
        """
        Alter the point stored in the location while preserving the labware.

        This returns a new Location and does not alter the current one. It
        should be used like

        .. code-block:: python

            >>> loc = Location(Point(1, 1, 1), None)
            >>> new_loc = loc.move(Point(1, 1, 1))
            >>>
            >>> # The new point is the old one plus the given offset.
            >>> assert new_loc.point == Point(2, 2, 2)  # True
            >>>
            >>> # The old point hasn't changed.
            >>> assert loc.point == Point(1, 1, 1)  # True

        """

        return Location(
            point=self.point + point,
            labware=self._given_labware,
            _meniscus_tracking=self._meniscus_tracking,
        )

    def __repr__(self) -> str:
        return f"Location(point={repr(self._point)}, labware={self._labware}, meniscus_tracking={self._meniscus_tracking})"


# TODO(mc, 2020-10-22): use MountType implementation for Mount
class Mount(enum.Enum):
    LEFT = enum.auto()
    RIGHT = enum.auto()
    EXTENSION = enum.auto()

    def __str__(self) -> str:
        return self.name

    @classmethod
    def ot2_mounts(cls) -> List["Mount"]:
        return [Mount.LEFT, Mount.RIGHT]

    @classmethod
    def string_to_mount(cls, mount: str) -> "Mount":
        if mount == "right":
            return cls.RIGHT
        elif mount == "left":
            return cls.LEFT
        else:
            return cls.EXTENSION


class MountType(str, enum.Enum):
    LEFT = "left"
    RIGHT = "right"
    EXTENSION = "extension"

    # TODO (spp, 2023-05-04): we should deprecate this and instead create an 'other_pipette_mount' method
    def other_mount(self) -> MountType:
        return MountType.LEFT if self is MountType.RIGHT else MountType.RIGHT

    def to_hw_mount(self) -> Mount:
        return {
            MountType.LEFT: Mount.LEFT,
            MountType.RIGHT: Mount.RIGHT,
            MountType.EXTENSION: Mount.EXTENSION,
        }[self]

    @staticmethod
    def from_hw_mount(mount: Mount) -> MountType:
        """Convert from Mount to MountType."""
        mount_map = {Mount.LEFT: MountType.LEFT, Mount.RIGHT: MountType.RIGHT}
        return mount_map[mount]


class PipetteMountType(enum.Enum):
    LEFT = "left"
    RIGHT = "right"
    COMBINED = "combined"  # added for 96-channel. Remove if not required

    def to_mount_type(self) -> MountType:
        return {
            PipetteMountType.LEFT: MountType.LEFT,
            PipetteMountType.RIGHT: MountType.RIGHT,
        }[self]


# What is this used for? Can we consolidate this into MountType?
# If not, can we change the 'GRIPPER' mount name to 'EXTENSION' so that it's
# consistent with all user-facing mount names?
class OT3MountType(str, enum.Enum):
    LEFT = "left"
    RIGHT = "right"
    GRIPPER = "gripper"


class AxisType(enum.Enum):
    X = "X"  # gantry
    Y = "Y"
    Z_L = "Z_L"  # left pipette mount Z
    Z_R = "Z_R"  # right pipette mount Z
    Z_G = "Z_G"  # gripper mount Z
    P_L = "P_L"  # left pipette plunger
    P_R = "P_R"  # right pipette plunger
    Q = "Q"  # hi-throughput pipette tiprack grab
    G = "G"  # gripper grab

    @classmethod
    def axis_for_mount(cls, mount: Mount) -> "AxisType":
        map_axis_to_mount = {
            Mount.LEFT: cls.Z_L,
            Mount.RIGHT: cls.Z_R,
            Mount.EXTENSION: cls.Z_G,
        }
        return map_axis_to_mount[mount]

    @classmethod
    def mount_for_axis(cls, axis: "AxisType") -> Mount:
        map_mount_to_axis = {
            cls.Z_L: Mount.LEFT,
            cls.Z_R: Mount.RIGHT,
            cls.Z_G: Mount.EXTENSION,
        }
        return map_mount_to_axis[axis]

    @classmethod
    def plunger_axis_for_mount(cls, mount: Mount) -> "AxisType":
        map_plunger_axis_mount = {Mount.LEFT: cls.P_L, Mount.RIGHT: cls.P_R}
        return map_plunger_axis_mount[mount]

    @classmethod
    def ot2_axes(cls) -> List["AxisType"]:
        return [
            AxisType.X,
            AxisType.Y,
            AxisType.Z_L,
            AxisType.Z_R,
            AxisType.P_L,
            AxisType.P_R,
        ]

    @classmethod
    def flex_gantry_axes(cls) -> List["AxisType"]:
        return [
            AxisType.X,
            AxisType.Y,
            AxisType.Z_L,
            AxisType.Z_R,
            AxisType.Z_G,
        ]

    @classmethod
    def ot2_gantry_axes(cls) -> List["AxisType"]:
        return [
            AxisType.X,
            AxisType.Y,
            AxisType.Z_L,
            AxisType.Z_R,
        ]


AxisMapType = Dict[AxisType, float]
StringAxisMap = Dict[str, float]


# TODO(mc, 2020-11-09): this makes sense in shared-data or other common
# model library
# https://github.com/Opentrons/opentrons/pull/6943#discussion_r519029833
class DeckSlotName(enum.Enum):
    """Deck slot identifiers."""

    # OT-2:
    SLOT_1 = "1"
    SLOT_2 = "2"
    SLOT_3 = "3"
    SLOT_4 = "4"
    SLOT_5 = "5"
    SLOT_6 = "6"
    SLOT_7 = "7"
    SLOT_8 = "8"
    SLOT_9 = "9"
    SLOT_10 = "10"
    SLOT_11 = "11"
    FIXED_TRASH = "12"

    # OT-3:
    SLOT_A1 = "A1"
    SLOT_A2 = "A2"
    SLOT_A3 = "A3"
    SLOT_B1 = "B1"
    SLOT_B2 = "B2"
    SLOT_B3 = "B3"
    SLOT_C1 = "C1"
    SLOT_C2 = "C2"
    SLOT_C3 = "C3"
    SLOT_D1 = "D1"
    SLOT_D2 = "D2"
    SLOT_D3 = "D3"

    @classmethod
    def from_primitive(cls, value: DeckLocation) -> DeckSlotName:
        str_val = str(value).upper()
        return cls(str_val)

    @classmethod
    def ot3_slots(cls) -> List["DeckSlotName"]:
        return [
            DeckSlotName.SLOT_A1,
            DeckSlotName.SLOT_A2,
            DeckSlotName.SLOT_A3,
            DeckSlotName.SLOT_B1,
            DeckSlotName.SLOT_B2,
            DeckSlotName.SLOT_B3,
            DeckSlotName.SLOT_C1,
            DeckSlotName.SLOT_C2,
            DeckSlotName.SLOT_C3,
            DeckSlotName.SLOT_D1,
            DeckSlotName.SLOT_D2,
            DeckSlotName.SLOT_D3,
        ]

    # TODO(mm, 2023-05-08):
    # Migrate callers off of this method. https://opentrons.atlassian.net/browse/RLAB-345
    def as_int(self) -> int:
        """Return this deck slot as an OT-2-style integer.

        For example, `SLOT_5.as_int()` and `SLOT_C2.as_int()` are both `5`.

        Deprecated:
            This will not make sense when the OT-3 has staging area slots.
        """
        return int(self.to_ot2_equivalent().value)

    def to_ot2_equivalent(self) -> DeckSlotName:
        """Return the OT-2 deck slot that's in the same place as this one.

        For example, `SLOT_C2.to_ot3_equivalent()` is `SLOT_5`.

        If this is already an OT-2 deck slot, returns itself.
        """
        return _ot3_to_ot2.get(self, self)

    def to_ot3_equivalent(self) -> DeckSlotName:
        """Return the OT-3 deck slot that's in the same place as this one.

        For example, `SLOT_5.to_ot3_equivalent()` is `SLOT_C2`.

        If this is already an OT-3 deck slot, returns itself.
        """
        return _ot2_to_ot3.get(self, self)

    def to_equivalent_for_robot_type(self, robot_type: RobotType) -> DeckSlotName:
        """Return the deck slot, for the given robot type, that's in the same place as this one.

        See `to_ot2_equivalent()` and `to_ot3_equivalent()`.
        """
        if robot_type == "OT-2 Standard":
            return self.to_ot2_equivalent()
        elif robot_type == "OT-3 Standard":
            return self.to_ot3_equivalent()

    @property
    def id(self) -> str:
        """This slot's unique ID, as it appears in the deck definition.

        This can be used to look up slot details in the deck definition.

        This is preferred over `.value` or `.__str__()` for explicitness.
        """
        return self.value

    def __str__(self) -> str:
        """Stringify to the unique ID.

        For explicitness, prefer using `.id` instead.
        """
        return self.id


_slot_equivalencies = [
    (DeckSlotName.SLOT_1, DeckSlotName.SLOT_D1),
    (DeckSlotName.SLOT_2, DeckSlotName.SLOT_D2),
    (DeckSlotName.SLOT_3, DeckSlotName.SLOT_D3),
    (DeckSlotName.SLOT_4, DeckSlotName.SLOT_C1),
    (DeckSlotName.SLOT_5, DeckSlotName.SLOT_C2),
    (DeckSlotName.SLOT_6, DeckSlotName.SLOT_C3),
    (DeckSlotName.SLOT_7, DeckSlotName.SLOT_B1),
    (DeckSlotName.SLOT_8, DeckSlotName.SLOT_B2),
    (DeckSlotName.SLOT_9, DeckSlotName.SLOT_B3),
    (DeckSlotName.SLOT_10, DeckSlotName.SLOT_A1),
    (DeckSlotName.SLOT_11, DeckSlotName.SLOT_A2),
    (DeckSlotName.FIXED_TRASH, DeckSlotName.SLOT_A3),
]

_ot2_to_ot3 = {ot2: ot3 for ot2, ot3 in _slot_equivalencies}
_ot3_to_ot2 = {ot3: ot2 for ot2, ot3 in _slot_equivalencies}


# TODO(jbl 11-17-2023) move this away from being an Enum and make this a NewType or something similar
class StagingSlotName(enum.Enum):
    """Staging slot identifiers."""

    SLOT_A4 = "A4"
    SLOT_B4 = "B4"
    SLOT_C4 = "C4"
    SLOT_D4 = "D4"

    @classmethod
    def from_primitive(cls, value: str) -> StagingSlotName:
        str_val = value.upper()
        return cls(str_val)

    @property
    def id(self) -> str:
        """This slot's unique ID, as it appears in the deck definition.

        This can be used to look up slot details in the deck definition.

        This is preferred over `.value` or `.__str__()` for explicitness.
        """
        return self.value

    def __str__(self) -> str:
        """Stringify to the unique ID.

        For explicitness, prefer using `.id` instead.
        """
        return self.id


class TransferTipPolicy(enum.Enum):
    ONCE = enum.auto()
    NEVER = enum.auto()
    ALWAYS = enum.auto()


DeckLocation = Union[int, str]
ALLOWED_PRIMARY_NOZZLES = ["A1", "H1", "A12", "H12"]


class NozzleConfigurationType(enum.Enum):
    """Short names for types of nozzle configurations.

    Represents the current nozzle configuration stored in a NozzleMap.
    """

    COLUMN = "COLUMN"
    ROW = "ROW"
    SINGLE = "SINGLE"
    FULL = "FULL"
    SUBRECT = "SUBRECT"


class NozzleMapInterface(Protocol):
    """
    A NozzleMap instance represents a specific configuration of active nozzles on a pipette.

    It exposes properties of the configuration like the configuration's front-right, front-left,
    back-left and starting nozzles as well as a map of all the nozzles active in the configuration.

    Because NozzleMaps represent configurations directly, the properties of the NozzleMap may not
    match the properties of the physical pipette. For instance, a NozzleMap for a single channel
    configuration of an 8-channel pipette - say, A1 only - will have its front left, front right,
    and active channels all be A1, while the physical configuration would have the front right
    channel be H1.
    """

    @property
    def starting_nozzle(self) -> str:
        """The nozzle that automated operations that count nozzles should start at."""
        ...

    @property
    def rows(self) -> dict[str, list[str]]:
        """A map of all the rows active in this configuration."""
        ...

    @property
    def columns(self) -> dict[str, list[str]]:
        """A map of all the columns active in this configuration."""
        ...

    @property
    def back_left(self) -> str:
        """The backest, leftest (i.e. back if it's a column, left if it's a row) nozzle of the configuration.

        Note: This is the value relevant for this particular configuration, and it may not represent the back left nozzle
        of the underlying physical pipette. For instance, the back-left nozzle of a configuration representing nozzles
        D7 to H12 of a 96-channel pipette is D7, which is not the back-left nozzle of the physical pipette (A1).
        """
        ...

    @property
    def configuration(self) -> NozzleConfigurationType:
        """The kind of configuration represented by this nozzle map."""
        ...

    @property
    def front_right(self) -> str:
        """The frontest, rightest (i.e. front if it's a column, right if it's a row) nozzle of the configuration.

        Note: This is the value relevant for this configuration, not the physical pipette. See the note on back_left.
        """
        ...

    @property
    def tip_count(self) -> int:
        """The total number of active nozzles in the configuration, and thus the number of tips that will be picked up."""
        ...

    @property
    def physical_nozzle_count(self) -> int:
        """The number of actual physical nozzles on the pipette, regardless of configuration."""
        ...

    @property
    def active_nozzles(self) -> list[str]:
        """An unstructured list of all nozzles active in the configuration."""
        ...



import logging
from typing import Optional

from opentrons.protocol_api import ProtocolContext
from opentrons.protocol_api._parameters import Parameters
from opentrons.protocols.execution.execute_python import exec_run
from opentrons.protocols.execution.json_dispatchers import (
    pipette_command_map,
    temperature_module_command_map,
    magnetic_module_command_map,
    thermocycler_module_command_map,
)
from opentrons.protocols.execution import execute_json_v4, execute_json_v3

from opentrons.protocols.types import PythonProtocol, Protocol
from opentrons.protocols.api_support.types import APIVersion

from opentrons.protocols.parameters.csv_parameter_interface import CSVParameter

MODULE_LOG = logging.getLogger(__name__)


def run_protocol(
    protocol: Protocol,
    context: ProtocolContext,
    run_time_parameters_with_overrides: Optional[Parameters] = None,
) -> None:
    """Run a protocol.

    :param protocol: The :py:class:`.protocols.types.Protocol` to execute
    :param context: The protocol context to use.
    :param run_time_parameters_with_overrides: Run time parameters defined in the protocol,
        updated with the run's RTP override values. When we are running either simulate
        or execute, this will be None (until RTP is supported in cli commands)
    """
    if isinstance(protocol, PythonProtocol):
        try:
            if protocol.api_level >= APIVersion(2, 0):
                exec_run(
                    proto=protocol,
                    context=context,
                    run_time_parameters_with_overrides=run_time_parameters_with_overrides,
                )
            else:
                raise RuntimeError(
                    f"Unsupported python API version: {protocol.api_level}"
                )
        except Exception:
            raise
        finally:
            if protocol.api_level >= APIVersion(2, 18):
                for parameter in context.params.get_all().values():
                    if isinstance(parameter, CSVParameter) and parameter.file_opened:
                        parameter.file.close()
    else:
        if protocol.contents["schemaVersion"] == 3:
            ins = execute_json_v3.load_pipettes_from_json(context, protocol.contents)
            lw = execute_json_v3.load_labware_from_json_defs(context, protocol.contents)
            execute_json_v3.dispatch_json(context, protocol.contents, ins, lw)
        elif protocol.contents["schemaVersion"] in [4, 5]:
            # reuse the v3 fns for loading labware and pipettes
            # b/c the v4 protocol has no changes for these keys
            ins = execute_json_v3.load_pipettes_from_json(context, protocol.contents)

            modules = execute_json_v4.load_modules_from_json(context, protocol.contents)

            lw = execute_json_v4.load_labware_from_json_defs(
                context, protocol.contents, modules
            )
            execute_json_v4.dispatch_json(
                context,
                protocol.contents,
                ins,
                lw,
                modules,
                pipette_command_map,
                magnetic_module_command_map,
                temperature_module_command_map,
                thermocycler_module_command_map,
            )
        else:
            raise RuntimeError(
                f"Unsupported JSON protocol schema: {protocol.schema_version}"
            )



from __future__ import annotations

from enum import Enum
import asyncio

from typing import Optional, Dict
from opentrons.drivers import utils
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.asyncio.communication import (
    AsyncResponseSerialConnection,
    UnhandledGcode,
)
from opentrons.drivers.heater_shaker.abstract import AbstractHeaterShakerDriver
from opentrons.drivers.types import Temperature, RPM, HeaterShakerLabwareLatchStatus


class GCODE(str, Enum):
    SET_RPM = "M3"
    GET_RPM = "M123"
    SET_TEMPERATURE = "M104"
    GET_TEMPERATURE = "M105"
    HOME = "G28"
    ENTER_BOOTLOADER = "dfu"
    GET_VERSION = "M115"
    OPEN_LABWARE_LATCH = "M242"
    CLOSE_LABWARE_LATCH = "M243"
    GET_LABWARE_LATCH_STATE = "M241"
    DEACTIVATE_HEATER = "M106"
    GET_RESET_REASON = "M114"


HS_BAUDRATE = 115200
DEFAULT_HS_TIMEOUT = 40
HS_COMMAND_TERMINATOR = "\n"
HS_ACK = "OK" + HS_COMMAND_TERMINATOR
HS_ERROR_KEYWORD = "err"
HS_ASYNC_ERROR_ACK = "async"
DEFAULT_COMMAND_RETRIES = 0


class HeaterShakerDriver(AbstractHeaterShakerDriver):
    @classmethod
    async def create(
        cls, port: str, loop: Optional[asyncio.AbstractEventLoop]
    ) -> HeaterShakerDriver:
        """
        Create a heater-shaker driver.

        Args:
            port: port or url of heater shaker
            loop: optional event loop

        Returns: driver
        """
        connection = await AsyncResponseSerialConnection.create(
            port=port,
            baud_rate=HS_BAUDRATE,
            timeout=DEFAULT_HS_TIMEOUT,
            ack=HS_ACK,
            loop=loop,
            error_keyword=HS_ERROR_KEYWORD,
            async_error_ack=HS_ASYNC_ERROR_ACK,
        )
        return cls(connection=connection)

    def __init__(self, connection: AsyncResponseSerialConnection) -> None:
        """
        Constructor

        Args:
            connection: SerialConnection to the heater-shaker
        """
        self._connection = connection

    async def connect(self) -> None:
        """Connect to heater-shaker"""
        await self._connection.open()

    async def disconnect(self) -> None:
        """Disconnect from heater-shaker"""
        await self._connection.close()

    async def is_connected(self) -> bool:
        """Check connection"""
        return await self._connection.is_open()

    async def open_labware_latch(self) -> None:
        """Send open-plate-lock command.

        Note: Labware latch is referred to as 'plate lock' in firmware.
        """
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.OPEN_LABWARE_LATCH
        )
        await self._connection.send_command(command=c, retries=DEFAULT_COMMAND_RETRIES)

    async def close_labware_latch(self) -> None:
        """Send close-plate-lock command.

        Note: Labware latch is referred to as 'plate lock' in firmware.
        """
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.CLOSE_LABWARE_LATCH
        )
        await self._connection.send_command(command=c, retries=DEFAULT_COMMAND_RETRIES)

    async def set_temperature(self, temperature: float) -> None:
        """Set temperature"""
        c = (
            CommandBuilder(terminator=HS_COMMAND_TERMINATOR)
            .add_gcode(gcode=GCODE.SET_TEMPERATURE)
            .add_float(
                prefix="S",
                value=temperature,
                precision=utils.HS_GCODE_ROUNDING_PRECISION,
            )
        )
        await self._connection.send_command(command=c, retries=DEFAULT_COMMAND_RETRIES)

    async def get_temperature(self) -> Temperature:
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_TEMPERATURE
        )
        response = await self._connection.send_command(
            command=c, retries=DEFAULT_COMMAND_RETRIES
        )
        return utils.parse_temperature_response(
            temperature_string=response, rounding_val=utils.HS_GCODE_ROUNDING_PRECISION
        )

    async def set_rpm(self, rpm: int) -> None:
        """Set RPM"""
        c = (
            CommandBuilder(terminator=HS_COMMAND_TERMINATOR)
            .add_gcode(gcode=GCODE.SET_RPM)
            .add_int(prefix="S", value=int(rpm))
        )
        await self._connection.send_command(command=c, retries=DEFAULT_COMMAND_RETRIES)

    async def get_rpm(self) -> RPM:
        """Get RPM"""
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_RPM
        )
        response = await self._connection.send_command(
            command=c, retries=DEFAULT_COMMAND_RETRIES
        )
        return utils.parse_rpm_response(rpm_string=response)

    async def get_labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        """Send get-labware-latch-status command.

        Note: Labware latch is referred to as 'plate lock' in firmware.
        """
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_LABWARE_LATCH_STATE
        )
        response = await self._connection.send_command(
            command=c, retries=DEFAULT_COMMAND_RETRIES
        )
        return utils.parse_labware_latch_status_response(status_string=response)

    async def home(self) -> None:
        """Send home command.

        Note: Homing also stops the shaking motion if applicable.
        """
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(gcode=GCODE.HOME)
        await self._connection.send_command(command=c, retries=DEFAULT_COMMAND_RETRIES)

    async def get_device_info(self) -> Dict[str, str]:
        """Send get-device-info command"""
        device_info = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_VERSION
        )
        response = await self._connection.send_command(
            command=device_info, retries=DEFAULT_COMMAND_RETRIES
        )

        reset_reason = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_RESET_REASON
        )
        try:
            await self._connection.send_command(
                command=reset_reason, retries=DEFAULT_COMMAND_RETRIES
            )
        except UnhandledGcode:
            pass

        return utils.parse_hs_device_information(device_info_string=response)

    async def enter_programming_mode(self) -> None:
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.ENTER_BOOTLOADER
        )
        await self._connection.send_dfu_command(command=c)
        await self._connection.close()

    async def deactivate_heater(self) -> None:
        """Send deactivate-heater command"""
        c = CommandBuilder(terminator=HS_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.DEACTIVATE_HEATER
        )
        await self._connection.send_command(command=c, retries=DEFAULT_COMMAND_RETRIES)
